from BinSizes import FreedmanDiaconisBinSize as bsc

def make_histogram_bins(bin_size_calculator, data, col_id):
  feature_vals = [row[col_id] for row in data]
  bin_range = bin_size_calculator(feature_vals)

  if bin_range==0.0:
    bin_range = 1.0

  data_tuples = list(enumerate(data)) # [(0,row), (1,row'), (2,row''), ... ]
  sorted_data_tuples = sorted(data_tuples, key=lambda tup: tup[1][col_id])

  max_val = max(data, key=lambda datum: datum[col_id])[col_id]
  min_val = min(data, key=lambda datum: datum[col_id])[col_id]

  index_bins = []
  val_ranges = []
  curr = min_val
  while curr <= max_val:
    index_bins.append([])
    val_ranges.append((curr, curr+bin_range))
    curr += bin_range

  for row_index, row in sorted_data_tuples:
    for bin_num, val_range in enumerate(val_ranges):
      if val_range[0] <= row[col_id] < val_range[1]:
        index_bins[bin_num].append(row_index)
        break

  index_bins = [b for b in index_bins if b]

  return index_bins


def test():
  data = [[i,0] for i in xrange(0, 100)]
  bins = make_histogram_bins(bsc, data, 0)
  print "make_histogram_bins -- no entries lost --", sum(len(row) for row in bins) == len(data)
  print "make_histogram_bins -- correct # of bins --", (len(bins) == 5)

  data = [[1]]*100
  bins = make_histogram_bins(bsc, data, 0)
  print "homogenous feature yields one bin? ", len(bins) == 1

  data = [[1]]*100 + [[2]]
  bins = make_histogram_bins(bsc, data, 0)
  print "bins being bucketed by value? -- ", len(bins) == 2



if __name__=="__main__":
  test()



from abc import ABCMeta, abstractmethod
import time

class AbstractModelFactory(object, metaclass=ABCMeta):
  def __init__(self, all_data, headers, response_header, name_prefix="",
               features_to_ignore=[], options={}):
    self.all_data = all_data
    self.headers = headers
    self.response_header = response_header
    self.features_to_ignore=features_to_ignore
    self.factory_name = "{}_{}".format(name_prefix, time.time()) if name_prefix else time.time()

    # All `options` should be consumed by the time the Abstract Factory is called.
    if options:
      raise Exception("Unknown ModelFactory options set: {}".format(options))

  @abstractmethod
  def build(self, train_set):
    pass



import operator
from collections import namedtuple, defaultdict
import csv
import ast
import os
import itertools


class Selector(namedtuple('Selector', 'var, op, val')):
  """
  Define a single rule condition
  """
  OPERATORS = {
    # discrete, nomial variables
    '==': operator.eq,
    '!=': operator.ne,
  
    # continous variables
    '<=': operator.le,
    '>=': operator.ge
  }

  def covers(self, x):
    return Selector.OPERATORS[self[1]](x[self[0]], self[2]) 
  
  def __str__(self):
    return str(self[0])+str(self[1])+str(self[2])

class Rule:
  def __init__(self, selectors=None, outcome_var=None, outcome_val=None,
         quality=None, influence_score=None, ID=None):
    self.selectors = selectors if selectors is not None else []
    self.outcome_var = outcome_var
    self.outcome_val = outcome_val
    self.quality = quality
    self.influence_score = influence_score
    self.ID = ID

  def calculate_influence_score(self, influence_scores):
    score = sum([influence_scores[s.var] for s in self.selectors])
    if self.influence_score is not None:
      assert(self.influence_score == score)
    else:
      self.influence_score = score

  def calculate_quality(self, data):
    covered_rows = []
    correct_covered_rows = []
    classes = set()
    for row in data:
      covers_row = True
      for selector in self.selectors:
        if not selector.covers(row):
          covers_row = False
      if covers_row:
        covered_rows.append(row)
        # check if outcome labels match
        if self.outcome_val == row[self.outcome_var]:
          correct_covered_rows.append(row)
      classes.add(row[self.outcome_var])

    num_covered = len(covered_rows)
    num_correct = len(correct_covered_rows)
    num_classes = len(classes)
    quality = (num_correct + 1.0) / (num_covered + num_classes)
    
    if self.quality is not None:
      assert(self.quality == quality)
    else:
      self.quality = quality

  def __str__(self):
    cond = " AND ".join([str(s) for s in self.selectors])
    outcome = "{}={}".format(self.outcome_var, self.outcome_val)
    return "IF {} THEN {}".format(cond, outcome)
    

"""
Format data 
"""
def get_data(data):
  data_dict = []
  headers = data[0]
  for row in data[1:]:
    row_dict = {}
    for i, element in enumerate(row):
      row_dict[headers[i]] = convert_ifnum(element)
    data_dict.append(row_dict)
  return data_dict

"""
Takes rules from the original rule list and converts them to Rule objects
"""
def is_int(string):
  try:
    int(string)
    return True
  except ValueError:
    return False

def is_float(string):
  try:
    float(string)
    return True
  except ValueError:
    return False

def convert_ifnum(string):
  if is_int(string):
    return int(string)
  if is_float(string):
    return float(string)
  return string

def parse_rule(rule):
  antecedent, consequent = rule.split(" THEN ")
  outcome_var, outcome_val = consequent.split("=")
  
  selectors = antecedent[3:].split(" AND ") # ignore "IF " in antecedent
  Selectors = []

  for selector in selectors:
    if "==" in selector:
      op = '=='
    elif "!=" in selector:
      op = '!='
    elif ">=" in selector:
      op = ">="
    elif "<=" in selector:
      op = "<="
    elif "TRUE" in selector:
      continue
    else:
      print(("Warning: unrecognized operation for " + selector))

    s_var, s_val = selector.split(op)
    s_val = convert_ifnum(s_val)
    Selectors.append(Selector(var=s_var, op=op, val=s_val))
    
  return Selectors, outcome_var, outcome_val
    

def get_rules_from_file(rulefile, data):
  reader = csv.DictReader(open(rulefile, "r"))
  rules = []
  for row in reader:
    rule_num = int(row["Label"])
    rule = row["Rules"]
    selectors, outcome_var, outcome_val = parse_rule(rule)
    influence_score = float(row["Score"])
    r = Rule(selectors=selectors, outcome_var=outcome_var, outcome_val=outcome_val, 
         influence_score=influence_score, ID=rule_num)
    r.calculate_quality(data)
    rules.append(r)
  return rules

"""
Get the possible values that original features are mapped to in the obscured version.
"""

def get_orig_to_obscured_map(orig_data, obscured_data):
  # dict mapping from attribute name to orig value to list of obscured values
  # "prior_count" -> { 3.0 -> [1.0, 2.0]}
  orig_to_obscured = {}

  # maps row num to the original value at an attribute
  # 2 -> {"prior_count" -> 1.0}
  rownum_to_origval = {}

  headers = obscured_data[0]
  for i, row in enumerate(orig_data):
    rownum_to_origval[i] = {}
    for j, val in enumerate(row):
      attr = headers[j]
      attr_val = convert_ifnum(val)
      rownum_to_origval[i][attr] = attr_val

      if attr not in orig_to_obscured:
        orig_to_obscured[attr] = {}
      if attr_val not in orig_to_obscured[attr]:
        orig_to_obscured[attr][attr_val] = []

  for i, row in enumerate(obscured_data[1:]):
    for j, val in enumerate(row):
      attr = headers[j]
      attr_val = convert_ifnum(val)

      if attr not in orig_to_obscured:
        print(("Warning: can't find original attribute to match this obscured attribte:" + attr))
      
      orig_val = rownum_to_origval[i][attr]

      if attr_val not in orig_to_obscured[attr][orig_val]:
        orig_to_obscured[attr][orig_val].append(attr_val)

  return orig_to_obscured

"""
Take a single rule and expand it to all the possible versions based on the obscured values.
"""

def generate_obscured_selectors(selector, obscured_vals, obscured_tag):
  obscured_selectors = [selector]
  obscured_var = selector.var + obscured_tag
  op = selector.op
  for obscured_val in obscured_vals:
    obscured_selectors.append(Selector(var=obscured_var, op=op, val=obscured_val))
  return obscured_selectors

def get_expanded_from_rule(original_rule, orig_to_obscured, data, influence_scores, obscured_tag):
  expanded_rules = [original_rule]
  possible_obscured_selectors = []
  for selector in original_rule.selectors:
    obscured_vals = orig_to_obscured[selector.var][selector.val]
    obscured_selectors = generate_obscured_selectors(selector, obscured_vals, obscured_tag)
    possible_obscured_selectors.append(obscured_selectors)

  product = itertools.product(*possible_obscured_selectors)
  for combo in product:
    outcome_var = original_rule.outcome_var
    outcome_val = original_rule.outcome_val
    ID = original_rule.ID
    expanded_rule = Rule(selectors=combo, outcome_var=outcome_var,
               outcome_val=outcome_val, ID=ID)
    expanded_rule.calculate_quality(data)
    expanded_rule.calculate_influence_score(influence_scores)
    expanded_rules.append(expanded_rule)
  return expanded_rules 

"""
Create a dictionary mapping original rule numbers to their expanded rule versions.
"""

def expand_rules(original_rules, orig_to_obscured, data, influence_scores, obscured_tag):
  expanded_rules_dict = {}
  for rule in original_rules:
    rule_num = rule.ID
    expanded_rules = get_expanded_from_rule(rule, orig_to_obscured, data,
                        influence_scores, obscured_tag)
    expanded_rules_dict[rule_num] = expanded_rules
  return expanded_rules_dict

"""
For each rule number, find the rule that is within epsilon quality of the original rule
  for that number and has the lowest influence score
"""

def select_best_obscured_rules(original_rules, expanded_rules_dict, by_original, epsilon):
  best_rules = []
  for original_rule in original_rules:
    ID = original_rule.ID
    # find best rule with quality within epsilon of the original rule's quality
    if by_original:
      best_rule = original_rule
      best_influence_score = original_rule.influence_score
      for expanded_rule in expanded_rules_dict[ID]:
        if (expanded_rule.quality + epsilon >= original_rule.quality and
          expanded_rule.influence_score < best_influence_score):
          best_rule = expanded_rule
          best_influence_score = expanded_rule.influence_score
      best_rules.append(best_rule)
    # find best rule with quality within epsilon of highest quality rule of bunch
    else:
      best_rule = None
      best_influence_score = 0.0
      best_quality = 0.0
      # find rule with the highest quality
      for rule in expanded_rules_dict[ID]:
        best_influence_score = rule.influence_score if rule.quality > best_quality else best_influence_score
        best_rule = rule if rule.quality > best_quality else best_rule
        best_quality = rule.quality if rule.quality > best_quality else best_quality
      for expanded_rule in expanded_rules_dict[ID]:
        if (expanded_rule.quality + epsilon >= best_quality and 
          expanded_rule.influence_score < best_influence_score):
          best_rule = expanded_rule
          best_influence_score = expanded_rule.influence_score
      best_rules.append(best_rule)
  return best_rules

"""
Find contexts of influence
"""
def find_contexts_of_influence(rules, obscured_tag):
  contexts_dict = defaultdict(set)
  for rule in rules:
    non_obscured_conds = frozenset([str(s) for s in rule.selectors if obscured_tag not in s.var])
    outcome = rule.outcome_var+'='+rule.outcome_val
    if non_obscured_conds:
      contexts_dict[outcome].add(non_obscured_conds)
  return contexts_dict

"""
Expand rules and find all contexts of influence
"""
def expand_and_find_contexts(original_csv, obscured_csv, merged_csv, rulesfile,
              influence_scores, obscured_tag, output_dir, by_original, epsilon):
  # Format data 
  data = get_data(merged_csv)

  # Convert rules form rule list into Rule objects
  original_rules = get_rules_from_file(rulesfile, data)
  
  # Get a mapping from the original data values to their obscured values
  orig_to_obscured = get_orig_to_obscured_map(original_csv, obscured_csv)
    
  print("Expanding rule list")
  # Expand each original rule and store them in a dictionary by the original rule's number
  expanded_rules_dict = expand_rules(original_rules, orig_to_obscured, data, influence_scores, obscured_tag)
  
  print("Finding best expanded rules")
  # Find the best expanded rule for each rule number
  best_rules = select_best_obscured_rules(original_rules, expanded_rules_dict, by_original, epsilon)

  # Find the contexts of influence for the best expanded rules
  contexts_of_influence = find_contexts_of_influence(best_rules, obscured_tag)
  
  # Write results to files 
  header = ["Label", "Rule", "Quality", "Influence"]

  print("Writing all expanded rules to file.")
  full_expanded_file = open("{}/full_expanded_rulelist.csv".format(output_dir), 'w')
  full_expanded_writer = csv.DictWriter(full_expanded_file, fieldnames=header)
  
  for rule_num in expanded_rules_dict:
    expanded_rules = expanded_rules_dict[rule_num]
    for rule in expanded_rules:
      rule_ID = rule.ID
      quality = rule.quality
      influence = rule.influence_score
      row_dict = {"Label": rule_ID, "Rule": str(rule), "Quality": quality, "Influence": influence}
      full_expanded_writer.writerow(row_dict)

  print("Writing best expanded rules results to file.")
  best_expanded_file = open("{}/best_expanded_rulelist.csv".format(output_dir), 'w')
  best_expanded_writer = csv.DictWriter(best_expanded_file, fieldnames=header)

  for rule in best_rules:
    rule_num = rule.ID
    quality = rule.quality
    influence = rule.influence_score
    row_dict = {"Label": rule_num, "Rule": str(rule), "Quality": quality, "Influence": influence}
    best_expanded_writer.writerow(row_dict)

  print("Writing contexts of influence results to file.")
  contexts_file = open("{}/contexts_of_influence.txt".format(output_dir), 'w')
  for outcome in contexts_of_influence:
    list_of_contexts = contexts_of_influence[outcome]
    contexts = " OR \n".join([" AND ".join(context) for context in list_of_contexts])
    contexts_file.write(outcome +': \n')
    contexts_file.write(contexts + '\n\n')

  return contexts_of_influence

import shutil, filecmp
def test():
  TMP_DIR = "tmp"
  if not os.path.exists(TMP_DIR):
    os.mkdir(TMP_DIR)
  
  test_orig_contents = [["Col1","Col2","Class"],
                        ["A","X",1],
                        ["A","X",1],
                        ["A","Y",0],
                        ["B","X",0],
                        ["B","Y",0]]

  test_obscured_contents = [["Col1", "Col2","Class"],
                            ["a","x",1],
                            ["a","x",1],
                            ["a","y",0],
                            ["b","x",0],
                            ["b","y",0]]


  test_merged_contents = [["Col1","Col1-tag","Col2","Col2-tag","Class"],
                          ["A","a","X","x",1],
                          ["A","a","X","x",1],
                          ["A","a","Y","y",0],
                          ["B","b","X","x",0],
                          ["B","b","Y","y",0]] 

  test_rulesfile = [["Label","Rule","Quality","Score"],
                    [0,"IF Col1!=A THEN Class=0",0.75,0.15],
                    [1,"IF Col2==X THEN Class=1",0.75,0.15],
                    [2,"IF TRUE THEN Class=0",.67,0]]

  test_full_expanded_rulelist = [[0,"IF Col1!=A THEN Class=0",0.25,0.15],
                                 [0,"IF Col1!=A THEN Class=0",0.25,0.15],
                                 [0,"IF Col1-notag!=a THEN Class=0",0.25,0.0],
                                 [1,"IF Col2==X THEN Class=1",0.2,0.15],
                                 [1,"IF Col2==X THEN Class=1",0.2,0.15],
                                 [1,"IF Col2-notag==x THEN Class=1",0.2,0.0],
                                 [2,"IF  THEN Class=0",0.14285714285714285,0.0],
                                 [2,"IF  THEN Class=0",0.14285714285714285,0]]

  test_best_expanded_rulelist = [[0,"IF Col1-notag!=a THEN Class=0",0.25,0.0],
                                 [1,"IF Col2-notag==x THEN Class=1",0.2,0.0],
                                 [2,"IF  THEN Class=0,0.14285714285714285",0.0]]

  test_influence_scores = {"ColA":.15, "ColB":.15, "ColA-tag":0, "ColB-tag":0}
  obscured_tag = "-tag"
  outputdir = TMP_DIR
  epsilon = .1

  test_files = [(TMP_DIR + "/test_orig", test_orig_contents),
                (TMP_DIR + "/test_obscured", test_obscured_contents), 
                (TMP_DIR + "/test_merged", test_merged_contents), 
                (TMP_DIR + "/test_rulesfile", tests_rulesfile),
                (TMP_DIR + "/test_full_rules", test_full_expanded_rulelist), 
                (TMP_DIR + "/test_best_rules", test_best_expanded_rulelist)]

  for test_file in test_files:
    with open(test_file[0], 'w') as csvf:
      f = csv.writer(csvf)
      for row in test_file[1]:
        f.writerow(row)

  contexts_of_influence = expand_and_find_contexts(original_csv, obscured_csv, merged_csv, rulesfile, influence_scores, obscured_tag, output_dir, by_original, epsilon)
  
  res_files = [(TMP_DIR+"/full_expanded_rulelist.csv"),
               (TMP_DIR+"/best_expanded_rulelist.csv")]

  assert(filecmp.cmp(res_files[0], test_file[4][0]))
  assert(filecmp.cmp(res_files[1], test_file[5][0]))

  shutil.rmtree(TMP_DIR)



from abc import ABCMeta, abstractmethod
import time

class AbstractModelVisitor(object):
  __metaclass__ = ABCMeta

  def __init__(self, model_name):
    self.model_name = "{}".format(time.time())

  @abstractmethod
  def test(self, test_set):
    pass



import csv
import sys

import matplotlib
matplotlib.use('Agg') # Set the back-end
import matplotlib.pyplot as plt

from os import listdir
from os.path import isfile, join

def graph_prediction_consistency(directory, output_image_file):
  only_files = [f for f in listdir(directory) if isfile(join(directory, f))]
  preds = ["{}/{}".format(directory, f) for f in only_files if ".predictions" in f]

  delim = ".audit.repaired_"

  ignored = ["original_train_data.predictions", "original_test_data.predictions"]

  file_groups = {}
  for pred in preds:
    if any(i in pred for i in ignored): continue
    feature = pred[len(directory)+1:pred.index(delim)] # Extract the feature name.
    if feature not in file_groups:
      file_groups[feature] = []
    file_groups[feature].append(pred)

  pred_groups = {}
  for feature, filenames in file_groups.items():
    pred_groups[feature] = []
    for filename in filenames:
      preds = load_pred_tups_from_predictions(filename)
      first_delim = filename.index(delim)+len(delim)
      second_delim = filename.index(".predictions")
      repair_level = float(filename[first_delim:second_delim])
      pred_groups[feature].append( (repair_level, preds) )
    pred_groups[feature].sort(key=lambda tup: tup[0]) # Sort by repair level.

  features = []
  y_axes = []
  for feature, pred_tups in pred_groups.items():
    orig = pred_tups[0][1]
    x_axis = [rep_level for rep_level, _ in pred_tups]
    y_axis = [similarity_to_original_preds(orig, tups) for _, tups in pred_tups]
    plt.plot(x_axis, y_axis, label=feature)

    features.append(feature)
    y_axes.append(y_axis)

  # Format and save the graph to an image file.
  plt.title("Similarity to Original Predictions")
  plt.axis([0,1,0,1.1]) # Make all the plots consistently sized.
  plt.xlabel("Repair Level")
  plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
  plt.savefig(output_image_file, bbox_inches='tight')
  plt.clf() # Clear the entire figure so future plots are empty.

  # Save the data used to generate that image file.
  with open(output_image_file + ".data", "w") as f:
    writer = csv.writer(f)
    headers = ["Repair Level"] + features
    writer.writerow(headers)
    for i, repair_level in enumerate(x_axis):
      writer.writerow([repair_level] + [y_vals[i] for y_vals in y_axes])


def load_pred_tups_from_predictions(filename):
  with open(filename) as f:
    reader = csv.reader(f)
    reader.next # Skip the headers.
    return [(r,p) for _,r,p in reader]

def similarity_to_original_preds(orig_pred_tups, new_pred_tups):
  matches = 0.0
  total = 0.0
  for orig, new in zip(orig_pred_tups, new_pred_tups):
    _,orig_pred = orig
    _,new_pred = new
    if orig_pred == new_pred:
      matches += 1
    total += 1

  return matches/total


if __name__=="__main__":
  if len(sys.argv) >= 2:
    directory = sys.argv[1]
    output_image = directory + "/similarity_to_original_predictions.png"
    graph_prediction_consistency(directory, output_image)
    print "Written to: {}".format(output_image)



import csv

def load_data():
  headers = ["Outcome", "X", "Y"]
  # Data from: https://github.com/jasonbaldridge/try-tf

  types = [str, float, float]
  with open("test_data/linear_data_train.csv") as f:
    reader = csv.reader(f)
    train = [[types[i](float(e)*100) for i, e in enumerate(row)] for row in reader]

  with open("test_data/linear_data_eval.csv") as f:
    reader = csv.reader(f)
    test = [[types[i](float(e)*100) for i, e in enumerate(row)] for row in reader]

  return headers, train, test

def test():
  headers, train, test = load_data()
  print("load_data -- unpacks correctly? -- ", (headers != None and train != None and test != None))

if __name__=="__main__":
  test()





class SparseList(list):
  def __init__(self, default=0, data=None):
    self.default = default
    self.vals = {}
    self.size = 0

    if data:
      self.extend(data)

  def __setitem__(self, index, value):
    if self.default != value:
      self.vals[index] = value
    self.size += 1

  def __len__(self):
    return self.size

  def __getitem__(self, index):
    if index in self.vals:
      return self.vals[index]
    else:
      return self.default

  def __repr__(self):
    return "<SparseList {}>".format(self.vals)

  def append(self, val):
    if self.default != val:
      self.vals[self.size] = val
    self.size += 1

  def extend(self, iterator):
    for val in iterator:
      if self.default != val:
        self.vals[self.size] = val
      self.size += 1

  def sort(self):
    values = sorted(self.vals.values())
    self.vals = {}
    old_size = self.size
    self.size = 0

    need_to_add_default = True
    for value in values:
      if need_to_add_default and self.default < value:
          self.size += (old_size - len(values))
          need_to_add_default = False

      if self.default != value:
        self.vals[self.size] = value
      self.size += 1

def audit_test():
  N=25000
  l = SparseList(default=0)
  for i in range(N):
    l.append(0)

  for i in range(N):
    l.append(i)

  for i in range(N):
    l.append(0)

  l.extend(range(N))

  l.sort()

  for i in range(4*N):
    l[i] # Call the `getter`

  print("Big SparseList size correct?", len(l) == 4*N)

def test():
  l = SparseList(default=0)
  l.append(0)
  l.append(0)
  l.append(-1)
  l.append(2)

  print("SparseList size correct?", len(l) == 4)
  print("SparseList accessed correctly?", l[0]==0 and l[2]==-1 and l[3] == 2)

  l.sort()
  print("Sorted SparseList size correct?", len(l) == 4)
  print("Sorted SparseList accessed correct?", l[0] == -1 and l[1]==0 and l[3] == 2)

  l.append(100)
  l.append(-100)
  l.sort()
  print("Resorted SparseList size correct?", len(l) == 6)
  print("Resorted SparseList accessed correct?",  l[0] == -100 and l[5]==100)


if __name__=="__main__":
  test()
  audit_test()




from abc import ABCMeta, abstractmethod
from AbstractModelFactory import AbstractModelFactory
from AbstractModelVisitor import AbstractModelVisitor

from collections import OrderedDict
import subprocess
import io
import csv
import os
import tempfile

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "weka.path"), 'r') as f:
  WEKA_PATH = f.readline()

if not os.path.isfile(WEKA_PATH):
  raise Exception("WEKA_PATH is not properly set!")

# Create the TMP_DIR if it does not already exist.
TMP_DIR = "tmp/"
if not os.path.exists(TMP_DIR):
  os.makedirs(TMP_DIR)

class AbstractWekaModelFactory(AbstractModelFactory):
  __metaclass__ = ABCMeta

  @abstractmethod
  def __init__(self, *args, **kwargs):
    super(AbstractWekaModelFactory, self).__init__(*args,**kwargs)
    self.train_command = ""
    self.model_visitor_type = None

  def build(self, train_set):

    # Prepare the ARFF file that will train the model.
    arff_types = get_arff_type_dict(self.headers, self.all_data)
    model_file = self.work_dir + "{}.model".format(self.verbose_factory_name)
    train_arff_file = model_file + ".train.arff"

    # Remove ignored features from the dataset.
    ignored_indices = [i for i,h in enumerate(self.headers) if h in self.features_to_ignore]
    actual_headers = [h for i,h in enumerate(self.headers) if i not in ignored_indices]
    train_set = [[elem for i,elem in enumerate(row) if i not in ignored_indices] for row in train_set]

    # Convert the final dataset to an ARFF for WEKA.
    list_to_arff_file(actual_headers,arff_types, train_set, train_arff_file)

    # Call WEKA to generate the model file.
    response_index = actual_headers.index(self.response_header)
    command = "java {} -t {} -d {} -p 0 -c {}".format(self.train_command, train_arff_file, model_file, response_index + 1)
    run_weka_command(command)

    return self.model_visitor_type(model_file, arff_types, response_index, actual_headers, ignored_indices)


class AbstractWekaModelVisitor(AbstractModelVisitor):

  def __init__(self, model_name, arff_types, response_index, headers, ignored_indices):
    self.model_name = model_name
    self.arff_types = arff_types
    self.response_index = response_index
    self.headers = headers
    self.ignored_indices = ignored_indices
    self.test_command = ""

  def test(self, test_set, test_name=""):
    # Remove ignored features from the test-set.
    test_set = [[elem for i,elem in enumerate(row) if i not in self.ignored_indices] for row in test_set]

    # Convert the final test-set to an ARFF for WEKA.
    test_arff_file = "{}.{}.test.arff".format(self.model_name, test_name)
    list_to_arff_file(self.headers, self.arff_types, test_set, test_arff_file)

    # Produce predictions for the test set.
    # Note: The "-c" option is 1-indexed by Weka
    results_path = "{}.out".format(test_arff_file)
    command = "java {} -T {} -l {} -p 0 -c {} 1> {}".format(self.test_command, test_arff_file, self.model_name, self.response_index+1, results_path)
    run_weka_command(command)

    # Read the output file.
    prediction_index = 2
    with open(results_path, "r") as f:
      raw_lines = f.readlines()[5:-1] # Discard the headers and ending line.
      raw_predictions = [line.split()[prediction_index] for line in raw_lines]
      predictions = [prediction.split(":")[1] for prediction in raw_predictions]

    return zip([row[self.response_index] for row in test_set], predictions)


def run_weka_command(command):
  set_path = "export CLASSPATH=$CLASSPATH:{}; ".format(WEKA_PATH)
  subprocess.check_output(set_path + command, shell=True)


def get_arff_type_dict(headers, data):
  values = {header:[row[i] for row in data] for i, header in enumerate(headers)}
  arff_type = OrderedDict()
  for header in headers:
    if all( map(lambda x: isinstance(x, float), values[header]) ):
      arff_type[header] = "numeric"
    elif all( map(lambda x: isinstance(x, bool), values[header]) ):
      arff_type[header] = [True,False]
    elif all( map(lambda x: isinstance(x, int), values[header]) ):
      arff_type[header] = "numeric"
    else:
      arff_type[header] = sorted(set(values[header])) # Categorical
  return arff_type


def list_to_arff_file(headers, arff_type_dict, data, arff_file_output):
  def arff_format(string):
    # Remove empty strings and quote strings with spaces.
    if string == "":
      string = "N/A"
    return '"{}"'.format(string) if " " in str(string) else string

  # Produce the relevant file headers for the ARFF.
  arff_header = "@relation BlackBoxAuditing\n"
  for header in headers:
    types = arff_type_dict[header]
    if type(types) == list:
      formatter = io.BytesIO()
      writer = csv.writer(formatter)
      unique_values = list(set(types))
      unique_values = [arff_format(val) for val in unique_values]
      writer.writerow(unique_values)
      formatted = formatter.getvalue().strip('\r\n')
      types = "{" + formatted + "}"
    header = header.replace(" ","_")
    types = types.replace('"""','"')
    arff_header += "@attribute {} {}\n".format(header, types)

  arff_header += "\n@data\n"

  # Write the data in a CSV-like format to avoid weird escaping issues.
  data_output = io.BytesIO()
  csv_writer = csv.writer(data_output)
  for row in data:
    row = [arff_format(val) for h, val in zip(headers,row) if h in arff_type_dict]
    csv_writer.writerow(row)

  # Dump everything into the intended ARFF file.
  with open(arff_file_output, "w") as f:
    output = data_output.getvalue().replace('"""','"')
    f.write(arff_header + output)



import random

def split_by_percent(data, train_percentage):
  train_indices = random.sample(range(len(data)), int(train_percentage*len(data)))
  train = [row for i,row in enumerate(data) if i in train_indices]
  test = [row for i,row in enumerate(data) if i not in train_indices]
  return train, test

def test():
  data = [[i] for i in range(100)]

  train,test = split_by_percent(data, 0.75)
  print "75:25 data split correct?", len(train) == 75 and len(test) == 25

  train,test = split_by_percent(data, 0.50)
  print "50:50 data split correct?", len(train) == 50 and len(test) == 50

if __name__=="__main__":
  test()




import csv
import random

train_filename = "test_data/adult.csv"
test_filename = "test_data/adult.test.csv"
max_entries = None
correct_types = [int, str, int, str, int, str, str, str,
                 str,str, int, int, int, str, str]

def load_data():
  with open(train_filename) as f:
    reader = csv.reader(f)
    train = [row for row in reader]
    headers = train.pop(0)

    if max_entries:
      train = random.sample(train, max_entries/2)

    for i, row in enumerate(train):
      for j, correct_type in enumerate(correct_types):
        train[i][j] = correct_type(row[j])

  with open(test_filename) as f:
    reader = csv.reader(f)
    test = [row for row in reader][1:] # Ignore headers.

    if max_entries:
      test = random.sample(test, max_entries/2)

    for i, row in enumerate(test):
      for j, correct_type in enumerate(correct_types):
        test[i][j] = correct_type(row[j])

  return headers, train, test


def test():
  headers, train, test = load_data()
  print "load_data unpacks correctly? -- ", (headers != None and train != None and test != None)

  gathered_types = []
  for i, header in enumerate(headers):
    if all( isinstance(row[i],float) for row in train + test ):
      gathered_types.append(float)
    elif all( isinstance(row[i],int) for row in train + test ):
      gathered_types.append(int)
    elif all( isinstance(row[i],str) for row in train + test ):
      gathered_types.append(str)

  print "load_data types are correct? -- ", gathered_types == correct_types
  print "load_data train size correct? -- ", len(train) == 32561
  print "load_data test size correct? -- ", len(test) == 16281

if __name__=="__main__":
  test()





"""
Development script for converting audit files into graphs.
"""

from audit_reading import graph_audits, graph_audit, rank_audit_files
from measurements import BCR, accuracy

from os import listdir
from os.path import isfile, join

import sys


def audit_directory(directory):
  only_files = [f for f in listdir(directory) if isfile(join(directory, f))]
  audits = ["{}/{}".format(directory, f) for f in only_files if f[-6:]==".audit"]

  measurers = [accuracy, BCR]
  for audit in audits:
    audit_image_filename = audit + ".png"
    graph_audit(audit, measurers, audit_image_filename)

  for measurer in measurers:
    ranked_graph_filename = "{}/{}.png".format(directory, measurer.__name__)
    graph_audits(audits, measurer, ranked_graph_filename)
    ranks = rank_audit_files(audits, measurer)
    print measurer.__name__, ranks


if __name__=="__main__":
  try:
    directory = sys.argv[1]
    audit_directory(directory)
  except:
    print "proper usage: make_graphs.py <directory/with/audits>"




import csv

from repairers.GeneralRepairer import Repairer

def kddtest(kdd):
  input_csv = 'test_data/RicciDataMod.csv'
  repair_level = 1.0
  protected_cols = ['Race']
  ignored = ['Position']

  with open(input_csv) as f:
    data = [line for line in csv.reader(f)]
    headers = data.pop(0)
    cols = [[row[i] for row in data] for i,col in enumerate(headers)]

    # Convert integer features to integers and float features to floats.
    for i, col in enumerate(cols):
      try:
        cols[i] = list(map(int, col))
      except ValueError:
        try:
          cols[i] = list(map(float, col))
        except ValueError:
          pass

    data = [[col[j] for col in cols] for j in range(len(data))]
    firstdata = data

  # Calculate the indices to repair by and to ignore.
  for protected in protected_cols:
    try:
      index_to_repair = headers.index(protected)
    except ValueError as e:
      raise Exception("Response header '{}' was not found in the following headers: {}".format(protected, headers))

    try:
      ignored_features = [headers.index(feature) for feature in ignored] if ignored else []
    except ValueError as e:
      raise Exception("One or more ignored-features were not found in the headers: {}".format(headers))

    repairer = Repairer(data, index_to_repair,
                        repair_level, kdd, features_to_ignore=ignored_features)


    # Repair the input data.
    data = repairer.repair(data)

    if not kdd:
      with open('repair_tests/RicciRepair1.0.KDDflag.csv') as f:
        correct_data = [line for line in csv.reader(f)]
    else:
      with open('repair_tests/RicciRepair1.0.noKDDflag.csv') as f:
        correct_data = [line for line in csv.reader(f)]

    # This is gross and I'm so sorry.
    strdata = []
    for line in data:
      for x in range(len(line)):
        line[x] = str(line[x])    
      strdata.append(line)

    if strdata == correct_data:
      return True
    else:
      return False

def test():
  print("Correct results for regular repair algorithm? -- ", kddtest(False))
  print("Correct results for KDD algorithm? -- ", kddtest(True))

if __name__ == "__main__": test()



from splitters import split_by_percent
from collections import OrderedDict

train_filename = "test_data/DRP_nature_train.arff"
test_filename = "test_data/DRP_nature_test.arff"
response = "outcome"
train_percentage = 2.0/3.0

def load_data():
  header_types = OrderedDict()
  data = []
  with open(train_filename) as f:
    for line in f:
      if "@attribute" in line:
        _, header, arff_type = line.split()
        header_types[header] = float if arff_type=="numeric" else str
      elif "@relation" in line or "@data" in line or line == "\n":
        pass
      else:
        row = line[:-1].split(",") #TODO: This is a naive way of splitting, captain.
        row = [header_types[h](v) for h,v in zip(header_types, row)]
        data.append(row)

  with open(test_filename) as f:
    for line in f:
      if "@attribute" not in line and "@relation" not in line and "@data" not in line and line != "\n":
        row = line[:-1].split(",") #TODO: This is a naive way of splitting, captain.
        row = [header_types[h](v) for h,v in zip(header_types, row)]
        data.append(row)

  headers = header_types.keys()

  # Translate the response into a binary.
  # index = headers.index(response)
  # translate = lambda row: row[:index] + ["1" if row[index] in "34" else "0"] + row[index+1:]
  # data = map(translate, data)

  train, test = split_by_percent(data, train_percentage)

  return headers, train, test


def test():
  headers, train, test = load_data()
  print "load_data unpacks correctly? -- ", (headers != None and train != None and test != None)

  correct_types = [float, float, str, float, str, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, str, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, float, str]

  gathered_types = []
  for i, header in enumerate(headers):
    if all( isinstance(row[i],float) for row in train + test ):
      gathered_types.append(float)
    elif all( isinstance(row[i],int) for row in train + test ):
      gathered_types.append(int)
    elif all( isinstance(row[i],str) for row in train + test ):
      gathered_types.append(str)

  print "load_data types are correct? -- ", gathered_types == correct_types
  print "load_data count is correct? -- ", 3955 == len(train) + len(test)


if __name__=="__main__":
  test()





from abc import ABCMeta, abstractmethod

class AbstractRepairer(object, metaclass=ABCMeta):
  """
  A Repairer object is capable of removing the correlations of a feature
  from a dataset at a specific `repair_level`.
  """

  def __init__(self, all_data, feature_to_repair, repair_level, kdd, features_to_ignore=[]):
    """
    all_data should be a list of rows (ie, a list of lists) composing the entire
    test and training dataset. Headers should not be included in data sets.

    feature_to_repair should be the index of the feature to repair. (ie, 0 to k)
    where k is the number of features in the dataset.

    repair_level should be a float between [0,1] representing the level of repair.

    features_to_ignore should be a list of feature indexes that should be ignored.
    """

    self.all_data = all_data
    self.feature_to_repair = feature_to_repair
    self.repair_level = repair_level
    self.kdd = kdd
    self.features_to_ignore = features_to_ignore

  @abstractmethod
  def repair(self, data_to_repair):
    """
    data_to_repair is the list of rows that actually should be repaired.
    """
    pass



import networkx as nx
from collections import defaultdict
import random

class CategoricalFeature:
  def __init__(self, data, name="no_name"):
    self.data = data
    self.name = name
    d1=defaultdict(int) #bin_data
    d2=defaultdict(int) #bin_index_dict
    d3=defaultdict(list) #bin_fulldata
    d4=defaultdict(int) #bin_index_dict_reverse
    d5=defaultdict(int) #category_count

    n = len(self.data)
    count = 0
    for i in range(0,n):
      obs = self.data[i]
      if obs in d2: pass  # if obs (i.e. category) is alreay a KEY in bin_index_data then don't do anything
      else:
        d2[obs] = count #bin_index_dict inits the KEY: category, with VALUE: count
        d4[count] = obs #bin_index_dict_reverse does the opposite
        count += 1
      bin_idx = d2[obs]
      d1[bin_idx] += 1 #add 1 to the obs category idex in bin_data
      d5[obs] += 1 #add 1 to the obs category NAME in category_count
      d3[bin_idx].append(i) #add obs to the list of obs with that category in bin_fulldata

    self.bin_data = d1
    self.category_count = d5
    self.num_bins = len(d1.items())
    self.bin_fulldata = d3
    self.bin_index_dict = d2
    self.bin_index_dict_reverse = d4


  def create_graph(self, count_generator): #creates graph given a CategoricalFeature object
    DG=nx.DiGraph() #using networkx package
    bin_list = self.bin_data.items()
    bin_index_dict_reverse = self.bin_index_dict_reverse
    k = self.num_bins
    DG.add_node('s')
    DG.add_node('t')
    for i in range(0, k): #lefthand side nodes have capacity = number of observations in category i
      DG.add_node(i)
      DG.add_edge('s', i, {'capacity' : bin_list[i][1], 'weight' : 0})
    for i in range(k, 2*k): #righthand side nodes have capacity = DESIRED number of observations in category i
      DG.add_node(i)
      cat = bin_index_dict_reverse[i-k]
      desired_count = count_generator(cat)
      DG.add_edge(i, 't', {'capacity' : desired_count, 'weight' : 0})
    #Add special node to hold overflow
    DG.add_node(2*k)
    DG.add_edge(2*k, 't', {'weight' : 0})
    for i in range(0, k):
      for j in range(k,2*k): #for each edge from a lefthand side node to a righhand side node:
        if (i+k)==j:  #IF they represent the same category, the edge weight is 0
          DG.add_edge(i, j, {'weight' : 0})
        else: #IF they represent different categories, the edge weight is 1
          DG.add_edge(i, j, {'weight' : 1})
      #THIS IS THE OVERFLOW NODE!!
      DG.add_edge(i, 2*k, {'weight' : 2})
    return DG

  def repair(self, DG): #new_feature = repair_feature(feature, create_graph(feature))
    mincostFlow = nx.max_flow_min_cost(DG, 's', 't') #max_flow_min_cost returns Dictionary of dictionaries. Keyed by nodes such that mincostFlow[u][v] is the flow edge (u,v)
    bin_dict = self.bin_fulldata
    index_dict = self.bin_index_dict_reverse
    size_data = len(self.data)
    repair_bin_dict = {}
    repair_data = [0]*size_data #initialize repaired data to be 0. If there are zero's after we fill it in the those observations belong in the overflow, "no category"
    k = self.num_bins
    overflow = 0
    for i in range(0,k): #for each lefthand side node i
      overflow += mincostFlow[i][2*k]
      for j in range(k, 2*k): #for each righthand side node j
        edgeflow = mincostFlow[i][j] #get the int (edgeflow) representing the amount of observations going from node i to j
        group = random.sample(bin_dict[i], int(edgeflow)) #randomly sample x (edgeflow) unique elements from the list of observations in that category.
        q=j-k #q is the category index for a righhand side node
        for elem in group: #for each element in the randomly selected group list
          bin_dict[i].remove(elem) #remove the element from the list of observation in that category
          repair_data[elem] = index_dict[q] #Mutate repair data at the index of the observation (elem) with its new category (it was 0) which is the category index for the righthand side node it flows to
        if q in repair_bin_dict: #if the category index is already keyed
          repair_bin_dict[q].extend(group) #extend the list of observations with a new list of observations in that category
        else:
          repair_bin_dict[q] = group #otherwise key that category index and set it's value as the group list in that category
    new_feature = CategoricalFeature(repair_data) #initialize our new_feature (repaired feature)
    new_feature.bin_fulldata = repair_bin_dict
    return [new_feature,overflow]


def test():
  random.seed(10)
  test_feature = CategoricalFeature(["A","B","C","D","D","D","C","B","A","C","B","A"])
  desired_count_dict = {"A": 1, "B": 2, "C": 2, "D": 3}
  desired_category_count = lambda category : desired_count_dict[category]
  DG = test_feature.create_graph(desired_category_count)
  [new_feature, overflow] = test_feature.repair(DG)
  edges = [
  (0, 8, {'weight': 2}), (0, 4, {'weight': 0}), (0, 5, {'weight': 1}), (0, 6, {'weight': 1}), (0, 7, {'weight': 1}),
  (1, 8, {'weight': 2}), (1, 4, {'weight': 1}), (1, 5, {'weight': 0}), (1, 6, {'weight': 1}), (1, 7, {'weight': 1}),
  (2, 8, {'weight': 2}), (2, 4, {'weight': 1}), (2, 5, {'weight': 1}), (2, 6, {'weight': 0}), (2, 7, {'weight': 1}),
  (3, 8, {'weight': 2}), (3, 4, {'weight': 1}), (3, 5, {'weight': 1}), (3, 6, {'weight': 1}), (3, 7, {'weight': 0}),
  (4, 't', {'capacity': 1, 'weight': 0}), (5, 't', {'capacity': 2, 'weight': 0}), (6, 't', {'capacity': 2, 'weight': 0}),
  (7, 't', {'capacity': 3, 'weight': 0}),
  (8, 't', {'weight': 0}),
  ('s', 0, {'capacity': 3, 'weight': 0}), ('s', 1, {'capacity': 3, 'weight': 0}),
  ('s', 2, {'capacity': 3, 'weight': 0}), ('s', 3, {'capacity': 3, 'weight': 0})]
  new_data = [0, 0, 'C', 'D', 'D', 'D', 'C', 'B', 'A', 0, 'B', 0]
  print "CategoricalFeature has correct number of categories?", 4 == test_feature.num_bins
  print "Directed Graph has correct edges and edge weights?", DG.edges(data=True) == edges
  print "mincostFlow has correct overflow?", overflow == 4
  print "mincostFlow has correct output data?", new_feature.data == new_data

if __name__=="__main__": test()



from .SVM_ModelFactory import ModelFactory as Weka_SVM
from .J48_ModelFactory import ModelFactory as Weka_DecisionTree
from .TensorFlowModelFactory import ModelFactory as TensorFlow
from .AbstractWekaModelFactory import AbstractWekaModelFactory, AbstractWekaModelVisitor



from abc import ABCMeta, abstractmethod
import time

class AbstractModelVisitor(object, metaclass=ABCMeta):
  def __init__(self, model_name):
    self.model_name = "{}".format(time.time())

  @abstractmethod
  def test(self, test_set):
    pass



from splitters import split_by_percent
import csv
from test_data import preloaded

def is_int(string):
  try:
    int(string)
    return True
  except ValueError:
    return False

def is_float(string):
  try:
    float(string)
    return True
  except ValueError:
    return False

def get_types(data,correct_types,empty_symbol):
  for i in range(len(correct_types)):
    row = 0
    while correct_types[i] is None:
      if data[row][i] == empty_symbol:
        row += 1
      else:
        if is_int(data[row][i]):
          correct_types[i] = int
        elif is_float[i] == float:
          correct_types[i] = float
        else:
          correct_types[i] = str

def load_data(data):
  if data not in preloaded:
    raise KeyError("{} is not an available dataset".format(data))
  if data == "DRP":
    return load_DRP(data)
  filename = preloaded[data]["filepath"]
  testdata = preloaded[data]["testdata"]
  correct_types = preloaded[data]["correct_types"]
  train_percentage = preloaded[data]["train_percentage"]
  response_header = preloaded[data]["response_header"]
  features_to_ignore = preloaded[data]["features_to_ignore"]
  with open(filename) as f:
    reader = csv.reader(f)
    data = [row for row in reader]
    headers = data.pop(0)

    for i, row in enumerate(data):
      for j, correct_type in enumerate(correct_types):
        data[i][j] = correct_type(row[j])

    if testdata is None:
      train, test = split_by_percent(data, train_percentage)
    else:
      train = data
      with open(testdata) as f:
        reader = csv.reader(f)
        test = [row for row in reader][1:] # Ignore headers.
  
        for i, row in enumerate(test):
          for j, correct_type in enumerate(correct_types):
            test[i][j] = correct_type(row[j])
  return headers, train, test, response_header, features_to_ignore


def load_DRP(data):
  train_filename = preloaded[data]["filepath"]
  test_filename = preloaded[data]["testdata"]
  train_percentage = preloaded[data]["train_percentage"]
  header = preloaded[data]["response_header"]
  features_to_ignore = preloaded[data]["features_to_ignore"]

  header_types = OrderedDict()
  data = []
  with open(train_filename) as f:
    for line in f:
      if "@attribute" in line:
        _, header, arff_type = line.split()
        header_types[header] = float if arff_type=="numeric" else str
      elif "@relation" in line or "@data" in line or line == "\n":
        pass
      else:
        row = line[:-1].split(",") #TODO: This is a naive way of splitting, captain.
        row = [header_types[h](v) for h,v in zip(header_types, row)]
        data.append(row)

  with open(test_filename) as f:
    for line in f:
      if "@attribute" not in line and "@relation" not in line and "@data" not in line and line != "\n":
        row = line[:-1].split(",") #TODO: This is a naive way of splitting, captain.
        row = [header_types[h](v) for h,v in zip(header_types, row)]
        data.append(row)

  headers = header_types.keys()

  train, test = split_by_percent(data, train_percentage)

  return headers, train, test, header, features_to_ignore


def load_from_file(datafile, testdata=None, correct_types=None, train_percentage=2.0/3.0,
                   response_header=None, features_to_ignore=None, missing_data_symbol=""):
  with open(datafile) as f:
    reader = csv.reader(f)
    data = [row for row in reader]
    headers = data.pop(0)

    # Set defaults in case they are not handed in as arguments
    if response_header is None:
      response_header = headers[-1]

    if features_to_ignore is None:
      features_to_ignore = []

    if correct_types is None:
	correct_types = get_types(data, [None]*len(headers), missing_data_symbol)
    
    for i, row in enumerate(data):
      for j, correct_type in enumerate(correct_types):
        data[i][j] = correct_type(row[j])

    if testdata is None:
      train, test = split_by_percent(data, train_percentage)
    else:
      train = data
      with open(testdata) as f:
        reader = csv.reader(f)
        test = [row for row in reader][1:] # Ignore headers.

        for i, row in enumerate(test):
          for j, correct_type in enumerate(correct_types):
            test[i][j] = correct_type(row[j])

  return headers, train, test, response_header, features_to_ignore



from splitters import split_by_percent
import csv

train_percentage = 2.0/3.0

def load_data():
  filename = ""  # file not publicly available
  with open(filename) as f:
    reader = csv.reader(f)
    data = [row for row in reader]
    headers = data.pop(0)

    correct_types = [str] * len(headers) # All categorical.

    for i, row in enumerate(data):
      for j, correct_type in enumerate(correct_types):
        data[i][j] = correct_type(row[j])

    train, test = split_by_percent(data, train_percentage)

  return headers, train, test




from splitters import split_by_percent
import random

train_percentage = 2.0/3.0

def load_data():
  N = 6000
  headers = ["Feature A (i)", "Feature B (2i)", "Feature C (-i)",
             "Constant Feature", "Random Feature", "Outcome"]

  data = [[i, 2*i, -i, 1, random.random(), "A"] for i in range(0,N/2)] + \
          [[i, 2*i, -i, 1, random.random(), "B"] for i in range(N/2,N)]

  train, test = split_by_percent(data, train_percentage)

  return headers, train, test

def test():
  headers, train, test = load_data()
  print "load_data -- unpacks correctly? -- ", (headers != None and train != None and test != None)

if __name__=="__main__":
  test()





# Parameters: feature_to_repair, response are both lists, groups and outcomes are tuples
# DI = Pr(C=1|X=0)/Pr(C=1|X=1)
def disparate_impact(list_of_triples, nonprotected_group, protected_group):
	# Assert len(feature_to_repair) == len(response), and groups and outcomes are tuples
	#SPECIFIC TO RECITIVISM DATA!!
	print("Calculating DI for: ", protected_group)
	feature_to_repair = []
	groups = {}
	response = []
	for triple in list_of_triples:
		#Actually, find for UKNOWN too if triple[0] != missing_group:
		groups[triple[0]] = 0
	for triple in list_of_triples:
		groups[triple[0]] += 1
		feature_to_repair.append(triple[0])
		response.append(int(triple[2]))
	group_list =[]
	for group in groups:
		group_list.append(group)
	total = len(feature_to_repair)
	#group_x white!
	group_x = [0]*total
	#group_y is not_white!
	group_y = [0]*total
	group_x_and_a = [0]*total	
	group_y_and_a = [0]*total	
	for i, group in enumerate(feature_to_repair):
		if group == nonprotected_group:
			group_x[i] = 1
			if response[i] == 0:
				group_x_and_a[i] = 1
		if group == protected_group:
			group_y[i] = 1 
			if response[i] == 0:
				group_y_and_a[i] = 1
	prob_x = sum(group_x) / float(total)
	prob_y = sum(group_y) / float(total)
	prob_x_and_a = sum(group_x_and_a) / float(total)
	prob_y_and_a = sum(group_y_and_a) / float(total)
	di = 0 
	if prob_x==0:
		prob_a_given_x=1
	else:
		prob_a_given_x = prob_x_and_a / prob_x
	if prob_y==0:
		prob_a_given_y=1
	else:
		prob_a_given_y = prob_y_and_a / prob_y
	if prob_a_given_x ==0:
		di= 1
	else:
		di = prob_a_given_y/prob_a_given_x
	print("Stats:", prob_x, prob_y, prob_x_and_a, prob_y_and_a, prob_a_given_x,prob_a_given_y,   di)
	return di


def disparate_impact2(feature_to_repair, response, groups, outcomes):
	# Assert len(feature_to_repair) == len(response), and groups and outcomes are tuples
	
	total = len(feature_to_repair)
	group_x = [0]*total
	group_y = [0]*total
	group_x_and_a = [0]*total	
	group_y_and_a = [0]*total	
	for i, group in enumerate(feature_to_repair):
		if group == groups[0]:
			group_x[i] = 1
			if response[i] == outcomes[0]:
				group_x_and_a[i] = 1
		elif group == groups[1]:
			group_y[i] = 1 
			if response[i] == outcomes[0]:
				group_y_and_a[i] = 1
	prob_x = sum(group_x) / float(total)
	prob_y = sum(group_y) / float(total)
	prob_x_and_a = sum(group_x_and_a) / float(total)
	prob_y_and_a = sum(group_y_and_a) / float(total)
	prob_a_given_x = prob_x_and_a / prob_x
	prob_a_given_y = prob_y_and_a / prob_y
	di = prob_a_given_y / prob_a_given_x
	return prob_a_given_y/prob_a_given_x

def test():
	feature_to_repair = ['W','W','B','B','W']
	response = [0,1,1,0,1]
	groups = ('W','B')
	outcomes = (1,0)
	di = disparate_impact2(feature_to_repair, response, groups, outcomes)
	di = round(di, 2)
	print("Disparate Impact correct?", di == 0.75) 

if __name__== "__main__":
  test()



from AbstractModelFactory import AbstractModelFactory
from AbstractModelVisitor import AbstractModelVisitor

import os
import numpy as np
import tensorflow as tf
import datetime

#TMP_DIR = "tmp/"
#CHECKPOINT_DIR = "tmp/tensorflow_checkpoints/"
#for directory in [TMP_DIR, CHECKPOINT_DIR]:
#  if not os.path.exists(directory):
#    os.makedirs(directory)


class ModelFactory(AbstractModelFactory):

  def __init__(self, *args, **kwargs):
    super(ModelFactory, self).__init__(*args, **kwargs)
    self.verbose_factory_name = "TensorFlow_Network"
    self.model_name = "Recidivism TensorFlow"

    self.num_epochs = 2000
    self.batch_size = 500

    self.response_index = self.headers.index(self.response_header)

    possible_values = set(row[self.response_index] for row in self.all_data)
    self.num_labels = len(possible_values)
    self.response_dict = {val:i for i,val in enumerate(possible_values)}

    self.hidden_layer_sizes = [100, 40] # If empty, no hidden layers are used.
    self.layer_types = [tf.nn.softmax, tf.nn.softmax, tf.nn.softmax]     # input layer, 4th Hidden Layer

  def build(self, train_set): #TODO: Add a features-to-ignore option.
    # In case the class is a string, translate it.
    translated_train_set = translate_response(self.response_index, train_set, self.response_dict)

    train_matrix, train_labels = list_to_tf_input(translated_train_set, self.response_index, self.num_labels)
    train_size, num_features = train_matrix.shape

    # Construct the layer architecture.
    x = tf.placeholder("float", shape=[None, num_features]) # Input
    y_ = tf.placeholder("float", shape=[None, self.num_labels]) # Output.

    layer_sizes = [num_features] + self.hidden_layer_sizes + [self.num_labels]
    # Generate a layer for the input and for each additional hidden layer.
    layers = [x] # Count the input as the first layer.
    for i in range(len(layer_sizes)-1):
      layer_size = layer_sizes[i]
      layer_type = self.layer_types[i]

      prev_layer = layers[-1]
      next_layer_size = layer_sizes[i+1]

      # Create a new layer with initially random weights and biases.
      W = tf.Variable(tf.random_normal([layer_size, next_layer_size]))
      b = tf.Variable(tf.random_normal([next_layer_size]))
      new_layer = layer_type(tf.matmul(prev_layer, W) + b)

      layers.append( new_layer )

    y = layers[-1]

    # Optimization.
    cross_entropy = -tf.reduce_sum(y_*tf.log(y))
    train_step = tf.train.GradientDescentOptimizer(0.01).minimize(cross_entropy)

    saver = tf.train.Saver()  # Defaults to saving all variables.

    # Create a local session to run this computation.
    with tf.Session() as tf_session:
      # For the test data, hold the entire dataset in one constant node.
      tf.initialize_all_variables().run()

      # Iterate and train.
      for step in range(self.num_epochs * train_size // self.batch_size):

        offset = (step * self.batch_size) % train_size
        batch_data = train_matrix[offset:(offset + self.batch_size), :]
        batch_labels = train_labels[offset:(offset + self.batch_size)]
        train_step.run(feed_dict={x: batch_data, y_: batch_labels})

        # Save the model file each step.
        saver.save(tf_session, CHECKPOINT_DIR + 'model.ckpt', global_step=step+1)

    return ModelVisitor(saver, self.response_index, self.num_labels, x, y_, y, self.response_dict)


class ModelVisitor(AbstractModelVisitor):

  def __init__(self, model_saver, response_index, num_labels, x, y_, y, response_dict):
    self.model_saver = model_saver
    self.model_name = "Recidivism TensorFlow Visitor"
    self.response_index = response_index
    self.num_labels = num_labels
    self.x = x
    self.y_ = y_
    self.y = y
    self.response_dict = response_dict

  def test(self, test_set, test_name=""):
    translated_test_set = translate_response(self.response_index, test_set, self.response_dict)

    test_matrix, test_labels = list_to_tf_input(translated_test_set, self.response_index, self.num_labels  )

    with tf.Session() as tf_session:
      ckpt = tf.train.get_checkpoint_state(CHECKPOINT_DIR)
      self.model_saver.restore(tf_session, ckpt.model_checkpoint_path)
      predictions = tf.argmax(self.y, 1).eval(feed_dict={self.x: test_matrix, self.y_:test_labels}, session=tf_session)

    predictions_dict = {i:key for key,i in list(self.response_dict.items())}
    predictions = [predictions_dict[pred] for pred in predictions]

    return list(zip([row[self.response_index] for row in test_set], predictions))

def list_to_tf_input(data, response_index, num_labels):
  data = expand_to_one_hot(data)
  matrix = np.matrix([row[:response_index] + row[response_index+1:] for row in data])

  labels = np.asarray([row[response_index] for row in data], dtype=np.uint8)
  labels_onehot = (np.arange(num_labels) == labels[:, None]).astype(np.float32)

  return matrix, labels_onehot

def translate_response(response_index, data_set, response_dict):
  translated_set = []
  for row in data_set:
    response = row[response_index]
    translation = response_dict[response]
    new_row = row[:response_index]+[translation]+row[response_index+1:]
    translated_set.append(new_row)
  return translated_set



def expand_to_one_hot(data,expand = True,use_alternative=False):
    header_dict = {'ALCABUS':0,'PRIRCAT':1,'TMSRVC':2,'SEX1':3,'RACE':4,'RELTYP':5,'age_1st_arrest':6,'DRUGAB':7,'Class':8,'RLAGE':9,'NFRCTNS':10}

    new_data = []
    for entry in data:
      temp = {}
    if expand == True:
      if entry[header_dict["SEX1"]] == "FEMALE":
        temp['female'] = 1
      else:
        temp['female'] = 0  

      if entry[header_dict["ALCABUS"]] == 'INMATE IS AN ALCOHOL ABUSER':
        temp['prior_alcohol_abuse'] = 1
      else:
        temp['prior_alcohol_abuse'] = 0

      if entry[header_dict['DRUGAB']] == 'INMATE IS A DRUG ABUSER':
        temp['prior_drug_abuse'] = 1
      else:
        temp['prior_drug_abuse'] = 0

      if entry[header_dict['NFRCTNS']] == 'INMATE HAS RECORD':
        temp['infraction_in_prison'] = 1
      else:
        temp['infraction_in_prison'] = 0

      race_cats = ['WHITE','BLACK','AMERICAN INDIAN/ALEUTIAN','ASIAN/PACIFIC ISLANDER','OTHER','UNKNOWN']

      for cat in race_cats:
        if entry[header_dict['RACE']] == cat:
          temp['race_'+cat] = 1
        else:
          temp['race_'+cat] = 0

        release_age_cats = ['14 TO 17 YEARS OLD','18 TO 24 YEARS OLD', '25 TO 29 YEARS OLD', \
        '30 TO 34 YEARS OLD','35 TO 39 YEARS OLD','40 TO 44 YEARS OLD','45 YEARS OLD AND OLDER']
      for cat in release_age_cats:
        if entry[header_dict['RLAGE']] == cat:
          temp['release_age_'+cat] = 1
        else:
          temp['release_age_'+cat] = 0

      time_served_cats = ['None','1 TO 6 MONTHS','13 TO 18 MONTHS','19 TO 24 MONTHS','25 TO 30 MONTHS', \
            '31 TO 36 MONTHS','37 TO 60 MONTHS','61 MONTHS AND HIGHER','7 TO 12 MONTHS']
      for cat in time_served_cats:
        if entry[header_dict['TMSRVC']] == cat:
          temp['time_served_'+cat] = 1
        else:
          temp['time_served_'+cat] = 0

      prior_arrest_cats = ['None','1 PRIOR ARREST','11 TO 15 PRIOR ARRESTS','16 TO HI PRIOR ARRESTS','2 PRIOR ARRESTS', \
        '3 PRIOR ARRESTS','4 PRIOR ARRESTS','5 PRIOR ARRESTS','6 PRIOR ARRESTS','7 TO 10 PRIOR ARRESTS']
      for cat in prior_arrest_cats:
        if entry[header_dict['PRIRCAT']] == cat:
          temp['prior_arrest_'+cat] = 1
        else:
          temp['prior_arrest_'+cat] = 0

      conditional_release =['PAROLE BOARD DECISION-SERVED NO MINIMUM','MANDATORY PAROLE RELEASE', 'PROBATION RELEASE-SHOCK PROBATION', \
            'OTHER CONDITIONAL RELEASE']
      unconditional_release = ['EXPIRATION OF SENTENCE','COMMUTATION-PARDON','RELEASE TO CUSTODY, DETAINER, OR WARRANT', \
            'OTHER UNCONDITIONAL RELEASE']
      other_release = ['NATURAL CAUSES','SUICIDE','HOMICIDE BY ANOTHER INMATE','OTHER HOMICIDE','EXECUTION','OTHER TYPE OF DEATH', \
            'TRANSFER','RELEASE ON APPEAL OR BOND','OTHER TYPE OF RELEASE','ESCAPE','ACCIDENTAL INJURY TO SELF','UNKNOWN']
      if entry[header_dict['RELTYP']] in conditional_release:
        temp['released_conditional'] = 1
        temp['released_unconditional'] = 0
        temp['released_other'] = 0
      elif entry[header_dict['RELTYP']] in unconditional_release:
        temp['released_conditional'] = 0
        temp['released_unconditional'] = 1
        temp['released_other'] = 0
      else:
        temp['released_conditional'] = 0
        temp['released_unconditional'] = 0
        temp['released_other'] = 1

      first_arrest_cats = ['UNDER 17','BETWEEN 18 AND 24','BETWEEN 25 AND 29','BETWEEN 30 AND 39','OVER 40']
      for cat in first_arrest_cats:
        if entry[header_dict['age_1st_arrest']] == cat:
          temp['age_first_arrest_'+cat] = 1
        else:
          temp['age_first_arrest_'+cat] = 0
    else:
      temp['SEX1'] = entry['SEX1']
      temp['RELTYP'] = entry['RELTYP']
      temp['PRIRCAT'] = entry['PRIRCAT']
      temp['ALCABUS'] = entry['ALCABUS']
      temp['DRUGAB'] = entry['DRUGAB']
      temp['RLAGE'] = entry['RLAGE']
      temp['TMSRVC'] = entry['TMSRVC']
      temp['NFRCTNS'] = entry['NFRCTNS']
      temp['RACE'] = entry['RACE']
      try:
        bdate = datetime.date(int(entry['YEAROB2']),int(entry['MNTHOB2']), int(entry['DAYOB2']))
        first_arrest = datetime.date(int(entry['A001YR']),int(entry['A001MO']),int(entry['A001DA']))
        first_arrest_age = first_arrest - bdate
        temp['age_1st_arrest'] = first_arrest_age.days
      except:
        temp['age_1st_arrest'] = 0

      new_data.append(temp)


    # convert from dictionary to list of lists
    fin = [[int(entry[key]) for key in list(entry.keys())] for entry in new_data]
    """
    with open("brandon_testing/test_"+str(time.clock())+".csv","w") as f:
    writer = csv.writer(f,delimiter=",")
    for row in fin:
        writer.writerow(row)
    """

    return fin





def test():
  # expand_to_one_hot assumes the data is the recidivism data, so these tests can't be run
  # test_list_to_tf_input()
  # test_basic_model()
  # test_categorical_model()
  pass

def test_list_to_tf_input():
  data = [[0,0],[0,1],[0,2]]
  tf_matrix, tf_onehot = list_to_tf_input(data, 1, 3)
  correct_matrix = [[0],[0],[0]]
  correct_onehot = [[1,0,0], [0,1,0], [0,0,1]]
  print("list_to_tf_input matrix correct? --",np.array_equal(tf_matrix, correct_matrix))
  print("list_to_tf_input onehot correct? --",np.array_equal(tf_onehot, correct_onehot))

def test_basic_model():
  headers = ["predictor 1", "predictor 2", "response"]
  response = "response"
  train_set = [[i,0,0] for i in range(1,50)] + [[0,i,1] for i in range(1,50)]
  test_set = [[i,0,0] for i in range(1,50)] + [[0,i,1] for i in range(1,50)]
  all_data = train_set + test_set

  factory = ModelFactory(all_data, headers, response, name_prefix="test")
  print("factory settings valid? -- ",len(factory.hidden_layer_sizes)+1 == len(factory.layer_types))

  model = factory.build(train_set)
  print("factory builds ModelVisitor? -- ", isinstance(model, ModelVisitor))

  predictions = model.test(test_set)
  resp_index = headers.index(response)
  intended_predictions = [(row[resp_index], row[resp_index]) for row in test_set]
  print("predicting numeric categories correctly? -- ", predictions == intended_predictions)

def test_categorical_model():
  headers = ["predictor 1", "predictor 2", "response"]
  response = "response"
  train_set = [[i,0,"A"] for i in range(1,50)] + [[0,i,"B"] for i in range(1,50)]
  test_set = [[i,0,"A"] for i in range(1,50)] + [[0,i,"C"] for i in range(1,50)]
  all_data = train_set + test_set

  factory = ModelFactory(all_data, headers, response, name_prefix="test")
  print("factory settings valid? -- ",len(factory.hidden_layer_sizes)+1 == len(factory.layer_types))

  model = factory.build(train_set)
  print("factory builds ModelVisitor? -- ", isinstance(model, ModelVisitor))

  predictions = model.test(test_set)
  resp_index = headers.index(response)
  intended_predictions = [(test_row[resp_index], train_row[resp_index]) for train_row, test_row in zip(train_set,test_set)]
  print("predicting string-categories correctly? -- ", predictions == intended_predictions)

if __name__=="__main__":
  test()



# NOTE: These settings and imports should be the only things that change
#       across experiments on different datasets and ML model types.

from BlackBoxAuditing.disparate_impact import disparate_impact
from BlackBoxAuditing.consistency_graph import *


from os import listdir
from os.path import isfile, join

def load_trip_from_predictions(filename):
  with open(filename) as f:
    reader = csv.reader(f)
    reader.__next__ # Skip the headers.
    return [(f,r,p) for f,r,p in reader]

def graph_disparate_impact_accuracy(directory, output_image_file):
  only_files = [f for f in listdir(directory) if isfile(join(directory, f))]
  preds = ["{}/{}".format(directory, f) for f in only_files if ".predictions" in f]

  delim = ".audit.repaired_"

  ignored = ["original_train_data.predictions", "original_test_data.predictions"]

  file_groups = {}
  for pred in preds:
    if any(i in pred for i in ignored): continue
    feature = pred[len(directory)+1:pred.index(delim)] # Extract the feature name.
    if feature not in file_groups:
      file_groups[feature] = []
    file_groups[feature].append(pred)

  pred_groups = {}
  for feature, filenames in list(file_groups.items()):
    pred_groups[feature] = []
    for filename in filenames:
      preds = load_trip_from_predictions(filename)
      first_delim = filename.index(delim)+len(delim)
      second_delim = filename.index(".predictions")
      repair_level = float(filename[first_delim:second_delim])
      pred_groups[feature].append( (repair_level, preds) )
    pred_groups[feature].sort(key=lambda tup: tup[0]) # Sort by repair level.

  protected_groups = ["WHITE","BLACK", "UNKNOWN", "ASIAN/PACIFIC ISLANDER", "AMERICAN INDIAN/ALEUTIAN"]
  unprotected_group = "WHITE"
  features = []
  y_axes = []
  for feature, pred_tups in list(pred_groups.items()):
    if feature == "RACE":
      for protected_group in protected_groups:
        #TODO clean this bit of code up, it's hacky right now
        real_pred_tups=[]
        for repair_level, triples in pred_tups:
          tups = []
          for triple in triples:
            tups.append((triple[1],triple[2]))
          real_pred_tups.append((repair_level,tups))
        
        orig = []
        for triple in pred_tups[0][1]:
          orig.append((triple[1],triple[2]))
        
        x_axis = [disparate_impact(triples[1:], unprotected_group, protected_group) for _,triples in pred_tups] 
        y_axis = [similarity_to_original_preds(orig, tups) for _, tups in real_pred_tups]
        plt.plot(x_axis, y_axis, label=protected_group)
        features.append(protected_group)
        y_axes.append(y_axis)

      # Format and save the graph to an image file.
      plt.title("Similarity to Original Predictions")
      plt.axis([.4,1.4,.7,1.1]) # Make all the plots consistently sized.
      plt.xlabel("Disparate Impact: Pr(Good Outcome given Race = X) / Pr(Good Outcome given Race = WHITE)")
      plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
      plt.savefig(output_image_file, bbox_inches='tight')
      plt.clf() # Clear the entire figure so future plots are empty.

      # Save the data used to generate that image file.
      with open(output_image_file + ".data", "w") as f:
        writer = csv.writer(f)
        headers = ["Disparate Impact"] + features
        writer.writerow(headers)
        for i, repair_level in enumerate(x_axis):
          writer.writerow([repair_level] + [y_vals[i] for y_vals in y_axes])

def graph_repair_level_disparate_impact(directory, output_image_file):
  only_files = [f for f in listdir(directory) if isfile(join(directory, f))]
  preds = ["{}/{}".format(directory, f) for f in only_files if ".predictions" in f]

  delim = ".audit.repaired_"

  ignored = ["original_train_data.predictions", "original_test_data.predictions"]

  file_groups = {}
  for pred in preds:
    if any(i in pred for i in ignored): continue
    feature = pred[len(directory)+1:pred.index(delim)] # Extract the feature name.
    if feature not in file_groups:
      file_groups[feature] = []
    file_groups[feature].append(pred)

  pred_groups = {}
  for feature, filenames in list(file_groups.items()):
    pred_groups[feature] = []
    for filename in filenames:
      preds = load_trip_from_predictions(filename)
      first_delim = filename.index(delim)+len(delim)
      second_delim = filename.index(".predictions")
      repair_level = float(filename[first_delim:second_delim])
      pred_groups[feature].append( (repair_level, preds) )
    pred_groups[feature].sort(key=lambda tup: tup[0]) # Sort by repair level.

  protected_groups = ["WHITE","BLACK", "UNKNOWN", "ASIAN/PACIFIC ISLANDER", "AMERICAN INDIAN/ALEUTIAN"]
  unprotected_group = "WHITE"
  features = []
  y_axes = []
  for feature, pred_tups in list(pred_groups.items()):
    if feature == "RACE":
      for protected_group in protected_groups:
        #TODO Figure out how to categorize feature values
        x_axis = [rep_level for rep_level,_ in pred_tups]
        #y_axis = [similarity_to_original_preds(orig, tups) for _, tups in pred_tups]
        y_axis = [disparate_impact(triples[1:], unprotected_group, protected_group) for _,triples in pred_tups]    
        #triples[0] = (Pre-Repaired Feature,Response,Prediction)
        #triples[1] = (WHITE,1,1)
        plt.plot(x_axis, y_axis, label=protected_group)
        features.append(protected_group)
        y_axes.append(y_axis)

      # Format and save the graph to an image file.
      plt.title("Disparate Impact: Pr(Good Outcome given Race = X) / Pr(Good Outcome given Race = WHITE)")
      plt.axis([0,1,0,1.8]) # Make all the plots consistently sized.
      plt.xlabel("Repair Level")
      plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
      plt.savefig(output_image_file, bbox_inches='tight')
      plt.clf() # Clear the entire figure so future plots are empty.

      # Save the data used to generate that image file.
      with open(output_image_file + ".data", "w") as f:
        writer = csv.writer(f)
        headers = ["Repair Level"] + features
        writer.writerow(headers)
        for i, repair_level in enumerate(x_axis):
          writer.writerow([repair_level] + [y_vals[i] for y_vals in y_axes])

if __name__=="__main__":
  graph_disparate_impact_accuracy("audits/1455586474.33", "disparate_impact_graphs/DI_Accuracy")
  graph_repair_level_disparate_impact("audits/1455586474.33", "disparate_impact_graphs/RepairLevel_DI")



import time

def expand_to_one_hot(data,expand = True):
    new_data = []
    for entry in data:
	temp = {}
	if expand == True:
	    if entry["SEX1"] == "FEMALE":
		temp['female'] = 1
	    else:
		temp['female'] = 0

	    if entry["ALCABUS"] == 'INMATE IS AN ALCOHOL ABUSER':
		temp['prior_alcohol_abuse'] = 1
	    else:
		temp['prior_alcohol_abuse'] = 0

	    if entry['DRUGAB'] == 'INMATE IS A DRUG ABUSER':
		temp['prior_drug_abuse'] = 1
	    else:
		temp['prior_drug_abuse'] = 0

	    if entry['NFRCTNS'] == 'INMATE HAS RECORD':
		temp['infraction_in_prison'] = 1
	    else:
		temp['infraction_in_prison'] = 0
    

	    release_age_cats = ['14 TO 17 YEARS OLD','18 TO 24 YEARS OLD', '25 TO 29 YEARS OLD', \
	    '30 TO 34 YEARS OLD','35 TO 39 YEARS OLD','40 TO 44 YEARS OLD','45 YEARS OLD AND OLDER']
	    for cat in release_age_cats:
		if entry['RLAGE'] == cat:
		    temp['release_age_'+cat] = 1
		else:
		    temp['release_age_'+cat] = 0
    
	    time_served_cats = ['None','1 TO 6 MONTHS','13 TO 18 MONTHS','19 TO 24 MONTHS','25 TO 30 MONTHS', \
			'31 TO 36 MONTHS','37 TO 60 MONTHS','61 MONTHS AND HIGHER','7 TO 12 MONTHS']
	    for cat in time_served_cats:
		if entry['TMSRVC'] == cat:
		    temp['time_served_'+cat] = 1
		else:
		    temp['time_served_'+cat] = 0

	    prior_arrest_cats = ['None','1 PRIOR ARREST','11 TO 15 PRIOR ARRESTS','16 TO HI PRIOR ARRESTS','2 PRIOR ARRESTS', \
		'3 PRIOR ARRESTS','4 PRIOR ARRESTS','5 PRIOR ARRESTS','6 PRIOR ARRESTS','7 TO 10 PRIOR ARRESTS']
	    for cat in prior_arrest_cats:
		if entry['PRIRCAT'] == cat:
		    temp['prior_arrest_'+cat] = 1
		else:
		    temp['prior_arrest_'+cat] = 0

	    conditional_release =['PAROLE BOARD DECISION-SERVED NO MINIMUM','MANDATORY PAROLE RELEASE', 'PROBATION RELEASE-SHOCK PROBATION', \
			'OTHER CONDITIONAL RELEASE']
	    unconditional_release = ['EXPIRATION OF SENTENCE','COMMUTATION-PARDON','RELEASE TO CUSTODY, DETAINER, OR WARRANT', \
			'OTHER UNCONDITIONAL RELEASE']
	    other_release = ['NATURAL CAUSES','SUICIDE','HOMICIDE BY ANOTHER INMATE','OTHER HOMICIDE','EXECUTION','OTHER TYPE OF DEATH', \
		    'TRANSFER','RELEASE ON APPEAL OR BOND','OTHER TYPE OF RELEASE','ESCAPE','ACCIDENTAL INJURY TO SELF','UNKNOWN']
	    if entry['RELTYP'] in conditional_release:
		temp['released_conditional'] = 1
		temp['released_unconditional'] = 0
		temp['released_other'] = 0
	    elif entry['RELTYP'] in unconditional_release:
		temp['released_conditional'] = 0
		temp['released_unconditional'] = 1
		temp['released_other'] = 0
	    else:
		temp['released_conditional'] = 0
		temp['released_unconditional'] = 0
		temp['released_other'] = 1

	    try:
		bdate = datetime.date(int(entry['YEAROB2']),int(entry['MNTHOB2']), int(entry['DAYOB2']))
		first_arrest = datetime.date(int(entry['A001YR']),int(entry['A001MO']),int(entry['A001DA']))
		first_arrest_age = first_arrest - bdate
		temp['age_1st_arrest'] = first_arrest_age.days
	    except:
		temp['age_1st_arrest'] = entry['age_1st_arrest']

	    # Add in the Y values
	    temp['Classarrests'] = entry['Classarrests']
	    temp['Classgeneral_violence'] = entry['Classgeneral_violence']
	    temp['Classfatal_violence'] = entry['Classfatal_violence']
	    temp['Classproperty'] = entry['Classproperty']
	    temp['Classsexual_violence'] = entry['Classsexual_violence']
	    temp['Classdrug'] = entry['Classdrug']
	else:
	    temp['SEX1'] = entry['SEX1']
	    temp['RELTYP'] = entry['RELTYP']
	    temp['PRIRCAT'] = entry['PRIRCAT']
	    temp['ALCABUS'] = entry['ALCABUS']
	    temp['DRUGAB'] = entry['DRUGAB']
	    temp['RLAGE'] = entry['RLAGE']
	    temp['TMSRVC'] = entry['TMSRVC']
	    temp['NFRCTNS'] = entry['NFRCTNS']
	    try:
		bdate = datetime.date(int(entry['YEAROB2']),int(entry['MNTHOB2']), int(entry['DAYOB2']))
		first_arrest = datetime.date(int(entry['A001YR']),int(entry['A001MO']),int(entry['A001DA']))
		first_arrest_age = first_arrest - bdate
		temp['age_1st_arrest'] = first_arrest_age.days
	    except:
		temp['age_1st_arrest'] = 0
    
	new_data.append(temp)


    return new_data



from AbstractRepairer import AbstractRepairer
import CategoricRepairer
from binning.Binner import make_histogram_bins
from binning.BinSizes import FreedmanDiaconisBinSize as bin_calculator
from calculators import get_median


class Repairer(AbstractRepairer):
  def __init__(self, *args, **kwargs):
    super(Repairer, self).__init__(*args, **kwargs)
    self.categoric_repairer = CategoricRepairer.Repairer(*args, **kwargs)

  def repair(self, data_to_repair):

    # Convert the "feature_to_repair" into a pseudo-categorical feature by
    # applying binning on that column.
    binned_data = [row[:] for row in data_to_repair]
    index_bins = make_histogram_bins(bin_calculator, data_to_repair, self.feature_to_repair)

    category_medians = {}
    for i, index_bin in enumerate(index_bins):
      bin_name = "BIN_{}".format(i) # IE, the "category" to replace numeric values.
      for j in index_bin:
        binned_data[j][self.feature_to_repair] = bin_name
      category_vals = [data_to_repair[j][self.feature_to_repair] for j in index_bin]
      category_medians[bin_name] = get_median(category_vals, self.kdd)

    repaired_data = self.categoric_repairer.repair(binned_data)

    # Replace the "feature_to_repair" column with the median numeric value.
    for i in xrange(len(repaired_data)):
      if self.repair_level > 0:
        rep_category = repaired_data[i][self.feature_to_repair]
        repaired_data[i][self.feature_to_repair] = category_medians[rep_category]
      else:
        repaired_data[i][self.feature_to_repair] = data_to_repair[i][self.feature_to_repair]
    return repaired_data


def test():
  test_sample()

def test_sample():
  data = [[float(i),float(i)*2, 1] for i in xrange(0, 150)]
  feature_to_repair = 0
  repairer = Repairer(data, feature_to_repair, 0.5, False)
  repaired_data = repairer.repair(data)
  print "repaired_data altered?", repaired_data != data

  repairer = Repairer(data, feature_to_repair, 0.0, False)
  repaired_data = repairer.repair(data)
  print "repaired_data unaltered for repair level=0?", repaired_data == data


if __name__=="__main__":
  test()




from BlackBoxAuditing.repairers.GeneralRepairer import Repairer
from BlackBoxAuditing.loggers import vprint
from BlackBoxAuditing.measurements import get_conf_matrix
from BlackBoxAuditing.model_factories.AbstractModelFactory import AbstractModelFactory
from BlackBoxAuditing.model_factories.AbstractModelVisitor import AbstractModelVisitor

from multiprocessing import Pool, cpu_count
import csv
import time
import os, shutil
import json
import gc

ENABLE_MULTIPROCESSING = False

# Used to share a copy of the dataset between multiprocessing processes.
shared_all = None
shared_train = None
shared_test = None

def _audit_worker(params):
  global shared_all
  global shared_train
  global shared_test

  model_or_factory, headers, ignored_features, feature_to_repair, repair_level, output_file, kdd, dump_all = params

  SAVE_REPAIRED_DATA = True if dump_all else False 
  SAVE_PREDICTION_DETAILS = True if dump_all else False

  index_to_repair = headers.index(feature_to_repair)

  repairer = Repairer(shared_all, index_to_repair,
                      repair_level, kdd, features_to_ignore=ignored_features)

  # Build a model on repaired training data if specified.
  if isinstance(model_or_factory, AbstractModelFactory):
    rep_train = repairer.repair(shared_train)
    model = model_or_factory.build(rep_train)

    # Log that this specific model was used for this repair level.
    with open(output_file + ".models.names.txt", "a") as f:
      f.write("{}: {}\n".format(repair_level, model.model_name))

    # Save the repaired version of the data if specified.
    if SAVE_REPAIRED_DATA:
      with open(output_file + ".train.repaired_{}.data".format(repair_level), "w") as f:
        writer = csv.writer(f)
        for row in [headers]+rep_train:
          writer.writerow(row)
  else:
    model = model_or_factory

  rep_test = repairer.repair(shared_test)

  test_name = "{}_{}".format(index_to_repair, repair_level)
  pred_tuples = model.test(rep_test, test_name=test_name)
  conf_table = get_conf_matrix(pred_tuples)
  
  # Save the repaired version of the data if specified.
  if SAVE_REPAIRED_DATA:
    with open(output_file + ".test.repaired_{}.data".format(repair_level), "w") as f:
      writer = csv.writer(f)
      for row in [headers]+rep_test:
        writer.writerow(row)
  
  repaired = output_file+".test.repaired_{}.data".format(repair_level)

  # Save the prediction_tuples and the original values of the features to repair.
  if SAVE_PREDICTION_DETAILS:
    with open(output_file + ".repaired_{}.predictions".format(repair_level), "w") as f:
      writer = csv.writer(f)
      file_headers = ["Pre-Repaired Feature", "Response", "Prediction"]
      writer.writerow(file_headers)
      for i, orig_row in enumerate(shared_test):
        row = [orig_row[index_to_repair], pred_tuples[i][0], pred_tuples[i][1]]
        writer.writerow(row)

  del repairer
  del rep_test
  gc.collect() 

  return repaired, (repair_level, conf_table)


class GradientFeatureAuditor(object):
  def __init__(self, model_or_factory, headers, train_set, test_set, kdd, repair_steps=10,
                features_to_ignore = None, features_to_audit=None, output_dir=None, dump_all=False):
    self.repair_steps = repair_steps
    self.model_or_factory = model_or_factory
    self.headers = headers
    self.features_to_ignore = features_to_ignore if features_to_ignore!=None else []
    self.OUTPUT_DIR = output_dir if output_dir!=None else "audits/{}".format(time.time())
    self.kdd = kdd
    self.dump_all = dump_all
    self.features_to_audit = features_to_audit
    self._rep_test = {}

    global shared_all
    global shared_train
    global shared_test

    shared_all = train_set + test_set
    shared_test = test_set
    shared_train = train_set

    # Set to `True` to allow the repaired data to be saved to a file.
    # Note: Be cautious when using this on large-sized datasets.

    # Create any output directories that don't exist.
    directory = self.OUTPUT_DIR
    if os.path.exists(directory):
      shutil.rmtree(directory)
    os.makedirs(directory)

  def audit_feature(self, feature_to_repair, output_file):
    repair_increase_per_step = 1.0/self.repair_steps
    repair_level = 0.0

    worker_params = []
    while repair_level <= 1.0:
      # Always save the full repair
      dump = self.dump_all if repair_level+repair_increase_per_step <= 1.0 else True
    
      call_params = (self.model_or_factory, self.headers, self.features_to_ignore, feature_to_repair, repair_level, output_file, self.kdd, dump)
      worker_params.append( call_params )
      repair_level += repair_increase_per_step

    if ENABLE_MULTIPROCESSING:
      pool = Pool(processes=cpu_count()/2 or 1, maxtasksperchild=1)
      conf_table_tuples = pool.map(_audit_worker, worker_params)
      pool.close()
      pool.join()
      del pool
    else:
      conf_table_tuples = [_audit_worker(params) for params in worker_params]

    conf_table_tuples.sort(key=lambda tuples: tuples[0])

    # Store location of full repaired data
    self._rep_test[feature_to_repair] = conf_table_tuples[-1][0]

    with open(output_file, "a") as f:
      f.write("GFA Audit for:{}\n".format(feature_to_repair))
      for repair_level, conf_table in conf_table_tuples:
        json_conf_table = json.dumps(conf_table)
        f.write("{}:{}\n".format(repair_level, json_conf_table))

  def audit(self, verbose=False):
    features_to_audit = [h for i, h in enumerate(self.headers) if i not in self.features_to_ignore] if self.features_to_audit is None else self.features_to_audit

    output_files = []
    for i, feature in enumerate(features_to_audit):
      message = "Auditing: '{}' ({}/{}).".format(feature,i+1,len(features_to_audit))
      vprint(message, verbose)

      cleaned_feature_name = feature.replace(".","_").replace(" ","_")
      output_file = "{}.audit".format(cleaned_feature_name)
      full_filepath = self.OUTPUT_DIR + "/" + output_file
      output_files.append(full_filepath)

      self.audit_feature(feature, full_filepath)

    audit_msg1 = "Audit file dump set to {}".format(self.dump_all)
    audit_msg2 = "All audit files have been saved." if self.dump_all else "Only mininal audit files have been saved."
    print("{}: {}".format(audit_msg1, audit_msg2))
    print("Audit files dumped to: {}.\n".format(self.OUTPUT_DIR))
    
    return output_files


class MockModel(AbstractModelVisitor):
  def test(self, test_set, arff_prefix="test", response_col=0, test_name=""):
    return [(entry[response_col], entry[response_col]) for entry in test_set]

def test():
  headers = ["response", "duplicate", "constant"]
  train = [[i,i,1] for i in range(100)]
  test = train[:] # Copy the training data.
  model = MockModel(test)
  repair_steps = 5
  gfa = GradientFeatureAuditor(model, headers, train, test, False,
                               repair_steps=repair_steps)
  output_files = gfa.audit()

  print("correct # of audit files produced? --", len(output_files) == len(train[0])) # The number of features.

  with open(output_files[0]) as f:
    print("correct # of lines per file? --", len(f.readlines()) == repair_steps+2) # +1 for the header-line and +1 for the level=0 step.

  files_not_empty = all(os.stat(f).st_size!=0 for f in output_files)
  print("all audit files not empty? --", files_not_empty)

  #TODO: Test the optional predictions and repaired output files.

if __name__=="__main__":
  test()



from BlackBoxAuditing.model_factories import Weka_SVM, Weka_DecisionTree, TensorFlow
from BlackBoxAuditing.loggers import vprint
from BlackBoxAuditing.GradientFeatureAuditor import GradientFeatureAuditor
from BlackBoxAuditing.audit_reading import graph_audit, graph_audits, rank_audit_files, group_audit_ranks
from BlackBoxAuditing.consistency_graph import graph_prediction_consistency
from BlackBoxAuditing.measurements import get_conf_matrix, accuracy, BCR
from BlackBoxAuditing.find_contexts import context_finder, load
from BlackBoxAuditing.data import load_data, load_from_file
from datetime import datetime
import csv
import os

class Auditor():
  def __init__(self):
    self.measurers = [accuracy, BCR]
    self.model_options = {}
    self.verbose = True
    self.REPAIR_STEPS = 10
    self.RETRAIN_MODEL_PER_REPAIR = False
    self.WRITE_ORIGINAL_PREDICTIONS = True  
    self.ModelFactory = Weka_SVM
    self.kdd = False
    self._audits_data = {}

  def __call__(self, data, output_dir=None, dump_all=False, features_to_audit=None):
    start_time = datetime.now()
  
    headers, train_set, test_set, response_header, features_to_ignore, correct_types = data

    self._audits_data = {"headers":headers, "train":train_set, "test":test_set, "response":response_header, "ignore":features_to_ignore, "types":correct_types, "full_audit":True if features_to_audit is None else False}
  
    """
     ModelFactories require a `build` method that accepts some training data
     with which to train a brand new model. This `build` method should output
     a Model object that has a `test` method -- which, when given test data
     in the same format as the training data, yields a confusion table detailing
     the correct and incorrect predictions of the model.
    """
  
    all_data = train_set + test_set
    model_factory = self.ModelFactory(all_data, headers, response_header,
                                      features_to_ignore=features_to_ignore,
                                      options=self.model_options)
  
    if not self.RETRAIN_MODEL_PER_REPAIR:
      vprint("Training initial model.",self.verbose)
      model = model_factory.build(train_set)
  
      # Check the quality of the initial model on verbose runs.
      if self.verbose:
        print("Calculating original model statistics on test data:")
        print("\tTraining Set:")
        train_pred_tuples = model.test(train_set)
        train_conf_matrix = get_conf_matrix(train_pred_tuples)
        print("\t\tConf-Matrix:", train_conf_matrix)
        for measurer in self.measurers:
          print("\t\t{}: {}".format(measurer.__name__, measurer(train_conf_matrix)))
  
        print("\tTesting Set:")
        test_pred_tuples = model.test(test_set)
        test_conf_matrix = get_conf_matrix(test_pred_tuples)
        print("\t\tConf-Matrix", test_conf_matrix)
        for measurer in self.measurers:
          print("\t\t{}: {}".format(measurer.__name__, measurer(test_conf_matrix)))
  
  
      model_or_factory = model
    else:
      model_or_factory = model_factory
  
    # Translate the headers into indexes for the auditor.
    audit_indices_to_ignore = [headers.index(f) for f in features_to_ignore]

    # Don't audit the response feature.
    audit_indices_to_ignore.append(headers.index(response_header))
  
    # Prepare the auditor.
    auditor = GradientFeatureAuditor(model_or_factory, headers, train_set, test_set,
                                     repair_steps=self.REPAIR_STEPS, kdd=self.kdd,
                                     features_to_ignore=audit_indices_to_ignore,
                                     features_to_audit=features_to_audit,
                                     output_dir=output_dir,dump_all=dump_all)
  
    # Perform the Gradient Feature Audit and dump the audit results into files.
    audit_filenames = auditor.audit(verbose=self.verbose)
 
    # Retrieve repaired data from audit
    self._audits_data["rep_test"] = auditor._rep_test
 
    ranked_features = []
    for measurer in self.measurers:
      vprint("Ranking audit files by {}.".format(measurer.__name__),self.verbose)
      ranked_graph_filename = "{}/{}.png".format(auditor.OUTPUT_DIR, measurer.__name__)
      ranks = rank_audit_files(audit_filenames, measurer)
      vprint("\t{}".format(ranks), self.verbose)
      ranked_features.append( (measurer, ranks) )

    end_time = datetime.now()
    
    # Store a summary of this experiment.
    summary = [
      "Audit Start Time: {}".format(start_time),
      "Audit End Time: {}".format(end_time),
      "Retrained Per Repair: {}".format(self.RETRAIN_MODEL_PER_REPAIR),
      "Model Factory ID: {}".format(model_factory.factory_name),
      "Model Type: {}".format(model_factory.verbose_factory_name),
      "Non-standard Model Options: {}".format(self.model_options),
      "Train Size: {}".format(len(train_set)),
      "Test Size: {}".format(len(test_set)),
      "Non-standard Ignored Features: {}".format(features_to_ignore),
      "Features: {}\n".format(headers)]

    # Print summary 
    for line in summary:
      print(line) 

    for ranker, ranks in ranked_features:
      print("Ranked Features by {}: {}".format(ranker.__name__, ranks))
      groups = group_audit_ranks(audit_filenames, ranker)
      print("\tApprox. Trend Groups: {}\n".format(groups))
      
      if ranker.__name__ == "accuracy":
        self._audits_data["ranks"] = ranks

    # Dump all experiment results if opted  
    if dump_all:
      vprint("Dumping original training data.", self.verbose)
      # Dump the train data to the log.
      train_dump = "{}/original_train_data".format(auditor.OUTPUT_DIR)
      with open(train_dump + ".csv", "w") as f:
        writer = csv.writer(f)
        writer.writerow(headers)
        for row in train_set:
          writer.writerow(row)
  
      if self.WRITE_ORIGINAL_PREDICTIONS:
        # Dump the predictions on the test data.
        with open(train_dump + ".predictions", "w") as f:
          writer = csv.writer(f)
          file_headers = ["Response", "Prediction"]
          writer.writerow(file_headers)
          for response, guess in train_pred_tuples:
            writer.writerow([response, guess])
  
      vprint("Dumping original testing data.", self.verbose)
      # Dump the train data to the log.
      test_dump = "{}/original_test_data".format(auditor.OUTPUT_DIR)
      with open(test_dump + ".csv", "w") as f:
        writer = csv.writer(f)
        writer.writerow(headers)
        for row in test_set:
          writer.writerow(row)
  
      if self.WRITE_ORIGINAL_PREDICTIONS:
        # Dump the predictions on the test data.
        with open(test_dump + ".predictions", "w") as f:
          writer = csv.writer(f)
          file_headers = ["Response", "Prediction"]
          writer.writerow(file_headers)
          for response, guess in test_pred_tuples:
            writer.writerow([response, guess])
 
      for measurer in self.measures:
         graph_audits(audit_filenames, measurer, ranked_graph_filename)     
 
      # Graph ://github.com/algofairness/BlackBoxAuditing.gitthe audit files.
      vprint("Graphing audit files.",self.verbose)
      for audit_filename in audit_filenames:
        audit_image_filename = audit_filename + ".png"
        graph_audit(audit_filename, self.measurers, audit_image_filename)
 
      # Store a graph of how many predictions change as features are repaired.
      vprint("Graphing prediction changes throughout repair.",self.verbose)
      output_image = auditor.OUTPUT_DIR + "/similarity_to_original_predictions.png"
      graph_prediction_consistency(auditor.OUTPUT_DIR, output_image)
  
      # Store a summary of this experiment to file.
      summary_file = "{}/summary.txt".format(auditor.OUTPUT_DIR)
      with open(summary_file, "w") as f:
        for line in summary:
          f.write(line+'\n')
  
        for ranker, ranks in ranked_features:
          f.write("Ranked Features by {}: {}\n".format(ranker.__name__, ranks))
          groups = group_audit_ranks(audit_filenames, ranker)
          f.write("\tApprox. Trend Groups: {}\n".format(groups))
  
      vprint("Summary file written to: {}\n".format(summary_file), self.verbose)

  def find_contexts(self, removed_attr, output_dir, beam_width=10, min_covered_examples=1, max_rule_length=5, by_original=True, epsilon=0.05):
    # retrive data from the audit
    audits_data = self._audits_data
    full_audit = audits_data["full_audit"]
    # Make sure a full audit was completed
    if not full_audit:
      raise RuntimeError("Only partial audit completed. Must run a full audit to call find_contexts.")

    orig_train = audits_data["train"]
    orig_test = audits_data["test"]
    obscured_test_data = audits_data["rep_test"][removed_attr]
    headers = audits_data["headers"]
    response_header = audits_data["response"]
    features_to_ignore = audits_data["ignore"]
    correct_types = audits_data["types"]
    obscured_tag = "-no"+removed_attr
    
    # Create directory to dump results
    if not os.path.exists(output_dir):
      os.makedirs(output_dir)

    # Extract influence scores
    ranks = audits_data["ranks"]
    influence_scores = {}
    for element in ranks:
      influence_scores[element[0]] = float(element[1])
      influence_scores[element[0]+obscured_tag] = 0.0

    # Get obscured data from file:
    obscured_test = []
    obscured_test_reader = csv.reader(open(obscured_test_data, 'r'))
    for row in obscured_test_reader:
      obscured_test.append(row)

    # load data from audit to prepare it for context finding process
    audit_params = (orig_train, orig_test, obscured_test, headers, response_header, features_to_ignore, correct_types, obscured_tag)

    orig_train_tab, orig_test_tab, merged_data = load(audit_params, output_dir)
   
    # run the context_finder 
    context_finder(orig_train, orig_test, obscured_test, orig_train_tab, orig_test_tab, merged_data, obscured_tag, output_dir, influence_scores, beam_width, min_covered_examples, max_rule_length, by_original, epsilon)

def main():
  # format data
  data = load_data("german")

  # set the auditor
  auditor = Auditor()
  auditor.model = Weka_SVM

  # call the auditor
  auditor(data, output_dir="german_audit_output", dump_all=False)

  auditor.find_contexts("age_cat",output_dir="german_context_output")

if __name__=="__main__":
  main()



import argparse
import time
import os
from BlackBoxAuditing.find_contexts.find_cn2_rules import CN2_learner
from BlackBoxAuditing.find_contexts.expand_and_find_contexts import expand_and_find_contexts


def context_finder(orig_train, orig_test, obscured_train, orig_train_tab, orig_test_tab, merged_data, obscured_tag, output_dir, influence_scores, beam_width, min_covered_examples, max_rule_length, by_original, epsilon):
  # Generate rule list for the original data using the CN2 algorithm
  rulesfile, accuracy, AUC = CN2_learner(orig_train_tab, orig_test_tab, output_dir, beam_width, min_covered_examples, max_rule_length, influence_scores)

  # Generate fully expanded rule list, store best expanded rule for each of the original rules,
  # and return contexts of discrimination
  contexts_of_influence = expand_and_find_contexts(orig_train, obscured_train, merged_data, rulesfile, influence_scores, obscured_tag, output_dir, by_original, epsilon)

  parsed_contexts = []
  for outcome in contexts_of_influence:
    list_of_contexts = contexts_of_influence[outcome]
    contexts = '\t'+" OR \n\t".join([" AND ".join(context) for context in list_of_contexts])
    parsed_contexts.append((outcome, contexts))

  # Store summary results:
  summary_info = ["\nCN2 Settings Used:",
                  "rules found for {}".format(orig_train_tab),
                  "beam_width: {}".format(beam_width),
                  "min_covered_examples: {}".format(min_covered_examples),
                  "max_rule_length: {}\n".format(max_rule_length),
                  "CN2 Model Evaluation:",
                  "Model tested on {}".format(orig_test_tab),
                  "Accuracy: {}".format(accuracy),
                  "AUC: {}\n".format(AUC)]

  # Print summary results
  for info_line in summary_info:
    print(info_line)

   
  print("\nContexts of influence found:")
  for parsed_context in parsed_contexts:
    print(parsed_context[0])
    print(parsed_context[1]+'\n\n')

    
  # Write results to summary file:
  summary = "{}/contexts.summary".format(output_dir)
  summary_file = open(summary, 'w')

  for info_line in summary_info:
    summary_file.write(info_line+'\n')

  summary_file.write("Contexts of influence found:\n")
  for parsed_context in parsed_contexts:
    summary_file.write(parsed_context[0]+':\n')
    summary_file.write(parsed_context[1]+'\n\n')

  print("Summary of Experiment written to {}".format(summary))



from itertools import product
from collections import defaultdict

from AbstractRepairer import AbstractRepairer
from CategoricalFeature import CategoricalFeature
from calculators import get_median
from SparseList import SparseList

import random
import math
from copy import deepcopy


class Repairer(AbstractRepairer):
  def repair(self, data_to_repair):
    num_cols = len(data_to_repair[0])
    col_ids = range(num_cols)
    
    # Get column type information
    col_types = ["Y"]*len(col_ids)
    for i, col in enumerate(col_ids):
      if i in self.features_to_ignore:
        col_types[i] = "I"
      elif i == self.feature_to_repair:
        col_types[i] = "X"

    col_type_dict = {col_id: col_type for col_id, col_type in zip(col_ids, col_types)}

    not_I_col_ids = filter(lambda x: col_type_dict[x] != "I", col_ids)
    
    if self.kdd:
      cols_to_repair = filter(lambda x: col_type_dict[x] == "Y", col_ids) 
    else:
      cols_to_repair = filter(lambda x: col_type_dict[x] in "YX", col_ids)
    
    # To prevent potential perils with user-provided column names, map them to safe column names
    safe_stratify_cols = [self.feature_to_repair]

    # Extract column values for each attribute in data
    # Begin byled code will usually be created in the same directory as the .py file. initializing keys and values in dictionary
    data_dict = {col_id: [] for col_id in col_ids}

    # Populate each attribute with its column values
    for row in data_to_repair:
      for i in col_ids:
        data_dict[i].append(row[i])


    repair_types = {}
    for col_id, values in data_dict.items():
      if all(isinstance(value, float) for value in values):
        repair_types[col_id] = float
      elif all(isinstance(value, int) for value in values):
        repair_types[col_id] = int
      else:
        repair_types[col_id] = str

    """
     Create unique value structures: When performing repairs, we choose median values. If repair is partial, then values will be modified to some intermediate value between the original and the median value. However, the partially repaired value will only be chosen out of values that exist in the data set.  This prevents choosing values that might not make any sense in the data's context.  To do this, for each column, we need to sort all unique values and create two data structures: a list of values, and a dict mapping values to their positions in that list. Example: There are unique_col_vals[col] = [1, 2, 5, 7, 10, 14, 20] in the column. A value 2 must be repaired to 14, but the user requests that data only be repaired by 50%. We do this by finding the value at the right index:
       index_lookup[col][2] = 1
       index_lookup[col][14] = 5
       this tells us that unique_col_vals[col][3] = 7 is 50% of the way from 2 to 14.
    """
    unique_col_vals = {}
    index_lookup = {}
    for col_id in not_I_col_ids:
      col_values = data_dict[col_id]
      # extract unique values from column and sort
      col_values = sorted(list(set(col_values)))
      unique_col_vals[col_id] = col_values
      # look up a value, get its position
      index_lookup[col_id] = {col_values[i]: i for i in range(len(col_values))}

    """
     Make a list of unique values per each stratified column.  Then make a list of combinations of stratified groups. Example: race and gender cols are stratified: [(white, female), (white, male), (black, female), (black, male)] The combinations are tuples because they can be hashed and used as dictionary keys.  From these, find the sizes of these groups.
    """
    unique_stratify_values = [unique_col_vals[i] for i in safe_stratify_cols]
    all_stratified_groups = list(product(*unique_stratify_values))
    # look up a stratified group, and get a list of indices corresponding to that group in the data
    stratified_group_indices = defaultdict(list)

    # Find the number of unique values for each strat-group, organized per column.
    val_sets = {group: {col_id:set() for col_id in cols_to_repair}
                                     for group in all_stratified_groups}
    for i, row in enumerate(data_to_repair):
      group = tuple(row[col] for col in safe_stratify_cols)
      for col_id in cols_to_repair:
        val_sets[group][col_id].add(row[col_id])

      # Also remember that this row pertains to this strat-group.
      stratified_group_indices[group].append(i)


    """
     Separate data by stratified group to perform repair on each Y column's values given that their corresponding protected attribute is a particular stratified group. We need to keep track of each Y column's values corresponding to each particular stratified group, as well as each value's index, so that when we repair the data, we can modify the correct value in the original data. Example: Supposing there is a Y column, "Score1", in which the 3rd and 5th scores, 70 and 90 respectively, belonged to black women, the data structure would look like: {("Black", "Woman"): {Score1: [(70,2),(90,4)]}}
    """
    stratified_group_data = {group: {} for group in all_stratified_groups}
    for group in all_stratified_groups:
      for col_id, col_dict in data_dict.items():
        # Get the indices at which each value occurs.
        indices = {}
        for i in stratified_group_indices[group]:
          value = col_dict[i]
          if value not in indices:
            indices[value] = []
          indices[value].append(i)

        stratified_col_values = [(occurs, val) for val, occurs in indices.items()]
        stratified_col_values.sort(key=lambda tup: tup[1])
        stratified_group_data[group][col_id] = stratified_col_values

    mode_feature_to_repair = get_mode(data_dict[self.feature_to_repair])

    # Repair Data and retrieve the results
    for col_id in cols_to_repair:
      # which bucket value we're repairing
      group_offsets = {group: 0 for group in all_stratified_groups}
      col = data_dict[col_id]

      num_quantiles = min(len(val_sets[group][col_id]) for group in all_stratified_groups)
      quantile_unit = 1.0/num_quantiles

      if repair_types[col_id] in {int, float}:
        for quantile in range(num_quantiles):
          median_at_quantiles = []
          indices_per_group = {}

          for group in all_stratified_groups:
            group_data_at_col = stratified_group_data[group][col_id]
            num_vals = len(group_data_at_col)
            offset = int(round(group_offsets[group]*num_vals))
            number_to_get = int(round((group_offsets[group] + quantile_unit)*num_vals) - offset)
            group_offsets[group] += quantile_unit

            if number_to_get > 0:

              # Get data at this quantile from this Y column such that stratified X = group
              offset_data = group_data_at_col[offset:offset+number_to_get]
              indices_per_group[group] = [i for val_indices, _ in offset_data for i in val_indices]
              values = sorted([float(val) for _, val in offset_data])

              # Find this group's median value at this quantile
              median_at_quantiles.append( get_median(values, self.kdd) )

          # Find the median value of all groups at this quantile (chosen from each group's medians)
          median = get_median(median_at_quantiles, self.kdd)
          median_val_pos = index_lookup[col_id][median]

          # Update values to repair the dataset.
          for group in all_stratified_groups:
            for index in indices_per_group[group]:
              original_value = col[index]

              current_val_pos = index_lookup[col_id][original_value]
              distance = median_val_pos - current_val_pos # distance between indices
              distance_to_repair = int(round(distance * self.repair_level))
              index_of_repair_value = current_val_pos + distance_to_repair
              repaired_value = unique_col_vals[col_id][index_of_repair_value]

              # Update data to repaired valued
              data_dict[col_id][index] = repaired_value

      #Categorical Repair is done below
      elif repair_types[col_id] in {str}:
        feature = CategoricalFeature(col)
        categories = feature.bin_index_dict.keys()

        group_features = get_group_data(all_stratified_groups, stratified_group_data, col_id)

        categories_count = get_categories_count(categories, all_stratified_groups, group_features)

        categories_count_norm = get_categories_count_norm(categories, all_stratified_groups, categories_count, group_features)

        median = get_median_per_category(categories, categories_count_norm)

        # Partially fill-out the generator functions to simplify later calls.
        dist_generator = lambda group_index, category : gen_desired_dist(group_index, category, col_id, median, self.repair_level, categories_count_norm, self.feature_to_repair, mode_feature_to_repair)

        count_generator = lambda group_index, group, category : gen_desired_count(group_index, group, category, median, group_features, self.repair_level, categories_count)

        group_features, overflow = flow_on_group_features(all_stratified_groups, group_features, count_generator)

        group_features, assigned_overflow, distribution = assign_overflow(all_stratified_groups, categories, overflow, group_features, dist_generator)

        # Return our repaired feature in the form of our original dataset
        for group in all_stratified_groups:
          indices = stratified_group_indices[group]
          for i, index in enumerate(indices):
            repaired_value = group_features[group].data[i]
            data_dict[col_id][index] = repaired_value

    # Replace stratified groups with their mode value, to remove it's information
    repaired_data = []
    for i, orig_row in enumerate(data_to_repair):
      new_row = [orig_row[j] if j not in cols_to_repair else data_dict[j][i] for j in col_ids]
      repaired_data.append(new_row)
    return repaired_data

def get_group_data(all_stratified_groups,stratified_group_data, col_id):
  group_features={}
  for group in all_stratified_groups:
    points = [(i, val) for indices, val in stratified_group_data[group][col_id] for i in indices]
    points = sorted(points, key=lambda x: x[0]) # Sort by index
    values = [value for _, value in points]

    # send values to CategoricalFeature object, which bins the data into categories
    group_features[group] = CategoricalFeature(values)
  return group_features


# Count the observations in each category. e.g. categories_count[1] = {'A':[1,2,3], 'B':[3,1,4]}, for column 1, category 'A' has 1 observation from group 'x', 2 from 'y', ect.
def get_categories_count(categories, all_stratified_groups, group_feature):
  # Grossness for speed efficiency. Don't worry, it make me sad, too.
  count_dict={cat: SparseList(data=(group_feature[group].category_count[cat] if cat in group_feature[group].category_count else 0 for group in all_stratified_groups)) for cat in categories}

  return count_dict

# Find the normalized count for each category, where normalized count is count divided by the number of people in that group
def get_categories_count_norm(categories, all_stratified_groups, count_dict, group_features):
  # Forgive me, Father, for I have sinned in bringing this monstrosity into being.
  norm = {cat: SparseList(data=(count_dict[cat][i] * (1.0/len(group_features[group].data)) if group_features[group].data else 0.0 for i,group in enumerate(all_stratified_groups))) for cat in categories}
  return norm

# Find the median normalized count for each category
def get_median_per_category(categories, categories_count_norm):
  return {cat: get_median(categories_count_norm[cat], False) for cat in categories}

# Generate the desired distribution and desired "count" for a given group-category-feature combination.
def gen_desired_dist(group_index, cat, col_id, median, repair_level, norm_counts, feature_to_remove, mode):
      if feature_to_remove == col_id:
        return 1 if cat==mode else (1-repair_level)*norm_counts[cat][group_index]
      else:
        return (1 - repair_level)*norm_counts[cat][group_index] + (repair_level*median[cat])

def gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count):
      med=median[category]
      size = len(group_features[group].data)
      # des-proportion = (1-lambda)*original-count  + (lambda)*median-count
      count = categories_count[category][group_index]
      des_count = math.floor(((1-repair_level)*count)+(repair_level)*med*size)
      return des_count

 # Run Max-flow to distribute as many observations to categories as possible. Overflow are those observations that are left over
def flow_on_group_features(all_stratified_groups, group_features, repair_generator):
  dict1= {}
  dict2={}
  for i, group in enumerate(all_stratified_groups):
    feature = group_features[group]
    count_generator = lambda category : repair_generator(i, group, category)

    # Create directed graph from nodes that supply the original counts to nodes that demand 
    # the desired counts, with a overflow node as total desired count is at most total original counts
    DG=feature.create_graph(count_generator)

    # Run max-flow, and record overflow count (total and per-group)
    new_feature,overflow = feature.repair(DG)
    dict2[group] = overflow

    # Update our original values with the values from max-flow, Note: still missing overflowed observations
    dict1[group] = new_feature

  return dict1, dict2

# Assign overflow observations to categories based on the group's desired distribution
def assign_overflow(all_stratified_groups, categories, overflow, group_features, repair_generator):
  feature = deepcopy(group_features)
  assigned_overflow = {}
  desired_dict_list = {}
  for group_index, group in enumerate(all_stratified_groups):
    # Calculate the category proportions.
    dist_generator = lambda cat: repair_generator(group_index, cat)
    cat_props = map(dist_generator,categories)

    if all(elem==0 for elem in cat_props): #TODO: Check that this is correct!
      cat_props = [1.0/len(cat_props)] * len(cat_props)
    s = float(sum(cat_props))
    cat_props = [elem/s for elem in cat_props]
    desired_dict_list[group] = cat_props
    assigned_overflow[group] = {}
    for i in range(int(overflow[group])):
      distribution_list = desired_dict_list[group]
      number = random.uniform(0, 1)
      cat_index = 0
      tally = 0
      for j in range(len(distribution_list)):
        value=distribution_list[j]
        if number < (tally+value):
          cat_index = j
          break
        tally += value
      assigned_overflow[group][i] = categories[cat_index]
    # Actually do the assignment
    count = 0
    for i, value in enumerate(group_features[group].data):
      if value == 0:
        (feature[group].data)[i] = assigned_overflow[group][count]
        count += 1
  return feature, assigned_overflow, desired_dict_list

def get_mode(values):
  counts = {}
  for value in values:
    counts[value] = 1 if value not in counts else counts[value]+1
  mode_tuple = max(counts.items(), key=lambda tup: tup[1])
  return mode_tuple[0]


def test():
  test_minimal()
  test_get_group_data()
  test_get_categories_count()
  test_get_categories_count_norm()
  test_get_median_per_category()
  test_gen_desired_count()
  test_gen_desired_dist()
  test_assign_overflow()
  test_categorical()
  test_numerical()
  test_repeated_values()

def test_repeated_values():
  all_data = [
  ["x","A"], ["x","B"], ["x","C"], ["x","D"], ["x","E"],
  ["y","F"], ["y","G"], ["y","H"], ["y","I"],
  ["z","J"], ["z","K"], ["z","L"], ["z","M"], ["z","N"], ["z","O"]]

  random.seed(10)

  repair_level=.5
  feature_to_repair = 0
  repairer = Repairer(all_data, feature_to_repair, repair_level, False)
  repaired_data=repairer.repair(all_data)

  correct_repaired_data = [
  ['x', 'E'], ['x', 'C'], ['x', 'C'], ['z', 'E'], ['x', 'B'], 
  ['y', 'I'], ['y', 'I'], ['y', 'G'], ['z', 'H'], 
  ['z', 'N'], ['z', 'N'], ['z', 'L'], ['z', 'K'], ['z', 'K'], ['z', 'K']]

  print "Test unique values -- .5 repaired_data altered?", repaired_data != all_data
  print "Test unique values -- .5 repaired_data correct?", repaired_data == correct_repaired_data

  all_data = [
  ["x","A"], ["x","A"], ["x","A"], ["x","A"], ["x","A"],
  ["y","A"], ["y","A"], ["y","A"], ["y","A"],
  ["z","A"], ["z","A"], ["z","A"], ["z","A"], ["z","A"], ["z","A"]]

  random.seed(10)

  repair_level=.5
  feature_to_repair = 0
  repairer = Repairer(all_data, feature_to_repair, repair_level, False)
  repaired_data=repairer.repair(all_data)

  correct_repaired_data = [
  ['x', 'A'], ['x', 'A'], ['x', 'A'], ['z', 'A'], ['x', 'A'], 
  ['y', 'A'], ['y', 'A'], ['y', 'A'], ['z', 'A'], 
  ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A']]


  print "Test repeated values -- .5 repaired_data altered?", repaired_data != all_data
  print "Test repeated values -- .5 repaired_data correct?", repaired_data == correct_repaired_data

def test_minimal():
  class_1 = [[float(i),"A"] for i in xrange(0, 100)]
  class_2 = [[float(i),"B"] for i in xrange(101, 200)] # Thus, "A" is mode class.
  data = class_1 + class_2

  feature_to_repair = 1
  repairer = Repairer(data, feature_to_repair, 1, False)
  repaired_data = repairer.repair(data)
  print "Minimal Dataset -- repaired_data altered?", repaired_data != data

  mode = get_mode([row[feature_to_repair] for row in data])
  print "Minimal Dataset -- mode is true mode?", mode=="A"
  print "Minimal Dataset -- mode value as feature_to_repair?", all(row[feature_to_repair] == mode for row in repaired_data)

def test_get_group_data():
  group_features = {}
  group_size = {}
  col_id = 1
  all_stratified_groups = [('y',),('z',)]
  stratified_group_data = {('y',): {1: [([4, 7, 5], 'A'),([3, 2, 6], 'B'), ([], 'C')]},\
                           ('z',): {1: [([9, 1, 10], 'A'), ([], 'B'), ([11], 'C')]}}
  group_features[col_id] = get_group_data(all_stratified_groups, stratified_group_data, col_id)

  print "Test get_group_data -- group features correct?", \
    [group_features[col_id][group].data for group in all_stratified_groups] == [['B','B','A','A','B','A'],['A', 'A', 'A', 'C']]
  #print "Test get_group_data -- group sizes correct?", group_size[col_id] == {('y',): 6, ('z',):4}

def test_get_categories_count():
  categories_count = {}
  categories = {1:['A','B','C','D']}
  all_stratified_groups = [('y',),('z',)]
  col_id = 1
  group_features = {1:{('y',): CategoricalFeature(['C','A','C','B','A','C']),\
                      ('z',): CategoricalFeature(['B','B','D','D'])}}\

  categories_count[col_id] = get_categories_count(categories[col_id], all_stratified_groups, group_features[col_id])
  #SparseList makes us test for correctness strangely
  print "Test get_categories_count -- category counts correct?",\
   categories_count[col_id]['A'][0] ==2 and categories_count[col_id]['C'][0] ==3 and categories_count[col_id]['B'][0] ==1\
   and categories_count[col_id]['B'][1] ==2 and categories_count[col_id]['D'][1] ==2

def test_get_categories_count_norm():
  categories_count_norm = {}
  categories = {1:['A','B']}
  all_stratified_groups = [('y',),('z',)]
  col_id = 1
  categories_count = {1: {'A':[4,0],'B':[16,0]}}
 
  #group_size = {1: {('y',): 20,('z',): 0}}
  group_features = {1:{('y',): CategoricalFeature(['A','A','A','A','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B']),
                        ('z',): CategoricalFeature([])}}
  categories_count_norm[col_id] = get_categories_count_norm(categories[col_id], all_stratified_groups, categories_count[col_id], group_features[col_id])
  #SparseList makes us test for correctness strangely
  print "Test get_categories_count_norm -- normalized category counts correct?",\
    categories_count_norm[col_id]['A'][0] == 0.2 and  categories_count_norm[col_id]['B'][0] == 0.8

def test_get_median_per_category():
  categories = {1:['A','B','C','D']}
  col_id =1
  categories_count_norm = {1:{'A':[0.25,0.0],'C':[0.3,0.0],'B':[0.4,0.0],  'D':[0.6,0.4]}}
  median = get_median_per_category(categories[col_id], categories_count_norm[col_id])
  print "Test get_median_per_category -- medians are correct?", median == {'A':0.0,'C':0.0,'B':0.0,  'D':0.4}

def test_gen_desired_count():
  #Case 1: feature with category with no values
  group_index = 0
  group = ('y',)
  median = {'A': 0.0 ,'B':0.0}
  group_features =  {('y',): CategoricalFeature(['A','A','A','A','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B']),
                        ('z',): CategoricalFeature([])}
  repair_level = .25
  categories_count = {'A':[4,0],'B':[16,0]}
  category = 'B'
  des_countB = gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)
  category = 'A'
  des_countA = gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)
  print "Test gen_desired_count -- desired count correct for feature with category with no values?", des_countB==12 and des_countA==3

  #Case 2: feature with regular categories
  group_index = 0
  group = ('y',)
  median = {'A': 0.2 ,'B':0.75}
  group_features =  {('y',): CategoricalFeature(['A','A','A','A', 'A','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B']),
                        ('z',): CategoricalFeature(['A','A','B','B','B','B','B','B','B','B'])}
  repair_level = 0.001
  categories_count = {'A':[5,2],'B':[15,8]}
  category = 'B'
  des_countB = gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)
  category = 'A'
  des_countA = gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)
  print "Test gen_desired_count -- desired count correct for standard feature?", des_countB==15 and des_countA==4 

  #Case 3: Repair feature after having repaired the other features on such feature 
  col_id = 0
  median = {'Y': 0.0 ,'Z':0.0}
  group_features =  {('y',): CategoricalFeature(['Y','Y','Y','Y']),
                       ('z',): CategoricalFeature(['Z','Z'])}
  repair_level = .5
  categories_count = {'Y':[4,0],'Z':[0,2]}
  mode_feature = 'Y' 
  group_index = 0
  group = ('y',)
  category = 'Y'
  des_countY = gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)
  group_index = 1
  group = ('z',)
  category = 'Z'
  des_countZ = gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)
  print "Test gen_desired_count -- desired count correct for mode category when repairing feature to remove?", des_countY==2 and des_countZ==1
  # If you are confused why desired count for Z is 2, it is beacuse our group_index is for group y 
  
def test_gen_desired_dist():
  #Case 1: feature with category with no values
  group_index = 0
  #group = ('y',)
  col_id = 1
  median = {'A': 0.0 ,'B':0.0}
  #group_features =  {('y',): CategoricalFeature(['A','A','A','A','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B']),
  #                      ('z',): CategoricalFeature([])}
  repair_level = .5
  categories_count_norm =  {'A':[0.2,0.0],'B':[0.8,0.0]}
  #categories_count = {'A':[4,0],'B':[16,0]}
  feature_to_remove = 0
  mode_feature = 'B' 
  category = 'B'
  des_distB = gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  category = 'A'
  des_distA = gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  print "Test gen_desired_dist -- desired distribution correct for feature with category with no values?", des_distB == .4 and des_distA == .1
  
  #Case 2: feature with regular categories
  group_index = 0
  #group = ('y',)
  median = {'A': 0.2 ,'B':0.75}
  #group_features =  {('y',): CategoricalFeature(['A','A','A','A', 'A','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B']),
  #                      ('z',): CategoricalFeature(['A','A','B','B','B','B','B','B','B','B'])}
  repair_level = 0.001
  categories_count_norm =  {'A':[0.25,0.2],'B':[0.75,0.8]}
  #categories_count = {'A':[5,2],'B':[15,8]}
  feature_to_remove = 0
  category = 'B'
  des_distB = gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  category = 'A'
  des_distA = gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  print "Test gen_desired_dist -- desired distribution correct for standard feature?", des_distB == .75 and des_distA == .24995 

  #Case 3: Repair feature after having repaired the other features on such feature 
  col_id = 0
  median = {'Y': 0.0 ,'Z':0.0}
  #group_features =  {('y',): CategoricalFeature(['Y','Y','Y','Y']),
  #                     ('z',): CategoricalFeature(['Z','Z'])}
  repair_level = .5
  categories_count_norm =  {'Y':[1.0,0.0],'Z':[0.0,1.0]}
  #categories_count = {'Y':[4,0],'Z':[0,2]}
  feature_to_remove = 0
  mode_feature = 'Y' 
  group_index = 0
  category = 'Y'
  des_distY = gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  group_index = 1
  category = 'Z'
  des_distZ = gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  print "Test gen_desired_dist -- desired distribution correct for mode category when repairing feature to remove?", des_distY==1 and des_distZ==.5

def test_assign_overflow():
  group_index = 0
  group = ('y',)
  category = 'B'
  col_id = 1
  median = {'A': 0.0 ,'B':0.0}
  group_features =  {('y',): CategoricalFeature(['A','A','A','A','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B','B']),
                      ('z',): CategoricalFeature([])}
  repair_level = 1
  categories_count_norm =  {'A':[0.2,0.0],'B':[0.8,0.0]}
  categories_count = {'A':[4,0],'B':[16,0]}
   
  feature_to_remove = 0
  mode_feature = 'B' 
  dist_generator = lambda group_index, category : gen_desired_dist(group_index, category, col_id, median, repair_level, categories_count_norm, feature_to_remove, mode_feature)
  count_generator = lambda group_index, group, category : gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count)

  random.seed(10)
  
  all_stratified_groups = [('y',),('z',)]
  categories = ['A','B']
  col_id = 1
  overflow = {('y',):2,('z',):2}
  group_features = {('y',):CategoricalFeature(['A','A','B','B',0,0]), ('z',): CategoricalFeature(['B',0,0])}

  feature, assigned_overflow, desired_dict_list = assign_overflow(all_stratified_groups, categories, overflow, group_features, dist_generator)
  print "Test assign_overflow -- updated group features correct?", \
   [feature[group].data for group in all_stratified_groups] ==[['A', 'A', 'B', 'B', 'B', 'A'], ['B', 'B', 'A']]
  print "Test assign_overflow -- assigned overflow correctly?", assigned_overflow == {('y',): {0: 'B', 1: 'A'}, ('z',): {0: 'B', 1: 'A'}}
  print "Test assign_overflow -- distribution correct?", desired_dict_list == {('y',): [0.5, 0.5], ('z',): [0.5, 0.5]}

def test_numerical():
  all_data = [ ["x", 1], ["x", 1], ["x", 2], ["x", 2], ["x", 3], ["x", 3],
               ["y", 1], ["y", 1], ["y", 2] ]
  feature_to_repair = 0
  repairer = Repairer(all_data, feature_to_repair, 1.0, False)
  repaired_data = repairer.repair(all_data)

  correct_repaired_data = [ ["x", 1], ["x", 1], ["x", 1], ["x", 1], ["x", 2], ["x", 2],
                            ["x", 1], ["x", 1], ["x", 2] ]

  print "Numerical minimal dataset -- full repaired_data correct?", repaired_data == correct_repaired_data


def test_categorical():
  all_data = [
  ["x","A"], ["x","A"], ["x","B"], ["x","B"], ["x","B"],
  ["y","A"], ["y","A"], ["y","A"], ["y","B"],
  ["z","A"], ["z","A"], ["z","A"], ["z","A"], ["z","A"], ["z","B"]]

  random.seed(10)

  repair_level=1
  feature_to_repair = 0
  repairer = Repairer(all_data, feature_to_repair, repair_level, False)
  repaired_data=repairer.repair(all_data)

  correct_repaired_data = [
  ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'B'], 
  ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'B'], 
  ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'B']]

  print "Categorical Minimal Dataset -- full repaired_data altered?", repaired_data != all_data
  print "Categorical Minimal Dataset -- full repaired_data correct?", repaired_data == correct_repaired_data

  repair_level=0.5
  feature_to_repair = 0
  repairer = Repairer(all_data, feature_to_repair, repair_level, False)
  part_repaired_data=repairer.repair(all_data)

  correct_part_repaired_data = [
  ['x', 'A'], ['x', 'A'], ['z', 'B'], ['z', 'B'], ['x', 'B'], 
  ['y', 'A'], ['y', 'A'], ['y', 'A'], ['y', 'B'], 
  ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'B']]

  print "Categorical Minimal Dataset -- partial (.5) repaired_data altered?", part_repaired_data != all_data
  print "Categorical Minimal Dataset -- partial (.5) repaired_data correct?", part_repaired_data == correct_part_repaired_data

  repair_level=0.2
  feature_to_repair = 0
  repairer = Repairer(all_data, feature_to_repair, repair_level, False)
  part_repaired_data=repairer.repair(all_data)

  correct_part2_repaired_data =  [
  ['x', 'A'], ['x', 'A'], ['x', 'B'], ['x', 'B'], ['z', 'B'], 
  ['y', 'A'], ['y', 'A'], ['y', 'A'], ['y', 'B'], 
  ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'A'], ['z', 'B']]

  print "Categorical Minimal Dataset -- partial (.2) repaired_data altered?", part_repaired_data != all_data
  print "Categorical Minimal Dataset -- partial (.2) repaired_data correct?", part_repaired_data == correct_part2_repaired_data


if __name__== "__main__":
  test()



from AbstractWekaModelFactory import AbstractWekaModelFactory, AbstractWekaModelVisitor, TMP_DIR

import os

class ModelFactory(AbstractWekaModelFactory):

  def __init__(self, *args, **kwargs):
    super(ModelFactory, self).__init__(*args,**kwargs)
    self.model_visitor_type = ModelVisitor

    self.verbose_factory_name = "J48_Decision_Tree"
    self.train_command = "weka.classifiers.trees.J48"

    self.work_dir = TMP_DIR + "{}_{}/".format(self.verbose_factory_name, self.factory_name)
    os.makedirs(self.work_dir)

class ModelVisitor(AbstractWekaModelVisitor):
  def __init__(self, *args, **kwargs):
    super(ModelVisitor, self).__init__(*args,**kwargs)
    self.test_command = "weka.classifiers.trees.J48"


def test():
  headers = ["predictor", "response"]
  A_set = [[i, "A"] for i in range(1,50)]
  B_set = [[i, "B"] for i in range(51,100)]
  C_set = [[i, "C"] for i in range(51,100)]
  train_set = A_set + B_set
  # Purposefully replace "B" with "C" in the test-set so that we *should* fail them.
  test_set = A_set + C_set
  all_data = train_set + test_set

  factory = ModelFactory(all_data, headers, "response", name_prefix="test")
  model = factory.build(train_set)
  print "factory builds ModelVisitor? -- ", isinstance(model, ModelVisitor)

  predictions = model.test(test_set)
  intended_predictions = [("A", "A")]*len(A_set) + [("C", "B")]*len(C_set)
  print "predicting correctly? -- ", predictions == intended_predictions

if __name__=="__main__":
  test()



import argparse
import csv

from repairers.GeneralRepairer import Repairer

def main():
  parser = argparse.ArgumentParser(description="Repair a CSV file.")
  parser.add_argument("input_csv", type=str,
                     help="The file on which to perform the repair.")
  parser.add_argument("output_csv", type=str,
                     help="The name to be used for the repaired CSV file.")
  parser.add_argument("repair_level", type=float,
                     help="The level at which the repair should be performed between 0.0 and 1.0.")
  parser.add_argument("kdd",type=str, 
                     help="Whether to repair with KDD.")

  parser.add_argument("-p", "--protected", type=str, nargs="+", required=True)
  parser.add_argument("-i", "--ignored", type=str, nargs="+")

  args = parser.parse_args()

  if args.kdd == "True":
    args.kdd = True
  else:
    args.kdd = False

  with open(args.input_csv) as f:
    data = [line for line in csv.reader(f)]
    headers = data.pop(0)
    cols = [[row[i] for row in data] for i,col in enumerate(headers)]

    # Convert integer features to integers and float features to floats.
    for i, col in enumerate(cols):
      try:
        cols[i] = map(int, col)
      except ValueError:
        try:
          cols[i] = map(float, col)
        except ValueError:
          pass

    data = [[col[j] for col in cols] for j in xrange(len(data))]
    firstdata = data

  # Calculte the indices to repair by and to ignore.
  for protected in args.protected:
    try:
      index_to_repair = headers.index(protected)
    except ValueError as e:
      raise Exception("Response header '{}' was not found in the following headers: {}".format(protected, headers))

    try:
      ignored_features = [headers.index(feature) for feature in args.ignored] if args.ignored else []
    except ValueError as e:
      raise Exception("One or more ignored-features were not found in the headers: {}".format(headers))

    repairer = Repairer(data, index_to_repair,
                        args.repair_level, args.kdd, features_to_ignore=ignored_features)


    # Repair the input data and write it to a CSV.
    data = repairer.repair(data)

  with open(args.output_csv, "wb") as f:
    writer = csv.writer(f)
    writer.writerow(headers)
    for row in data:
      writer.writerow(row)

if __name__ == "__main__":
  main()



from model_factories import Weka_SVM, Weka_DecisionTree, TensorFlow
from loggers import vprint
from GradientFeatureAuditor import GradientFeatureAuditor
from audit_reading import graph_audit, graph_audits, rank_audit_files, group_audit_ranks
from consistency_graph import graph_prediction_consistency
from measurements import get_conf_matrix, accuracy, BCR
from datetime import datetime
from data import load_data, load_from_file
import csv

class Auditor():
  def __init__(self):
    self.measurers = [accuracy, BCR]
    self.model_options = {}
    self.verbose = True
    self.REPAIR_STEPS = 10
    self.RETRAIN_MODEL_PER_REPAIR = False
    self.WRITE_ORIGINAL_PREDICTIONS = True  
    self.ModelFactory = Weka_SVM
    self.kdd = False

  def __call__(self, data, output_dir=None, dump_all=False, features_to_audit=None):
    start_time = datetime.now()
  
    headers, train_set, test_set, response_header, features_to_ignore = data
  
    """
     ModelFactories require a `build` method that accepts some training data
     with which to train a brand new model. This `build` method should output
     a Model object that has a `test` method -- which, when given test data
     in the same format as the training data, yields a confusion table detailing
     the correct and incorrect predictions of the model.
    """
  
    all_data = train_set + test_set
    model_factory = self.ModelFactory(all_data, headers, response_header,
                                      features_to_ignore=features_to_ignore,
                                      options=self.model_options)
  
    if not self.RETRAIN_MODEL_PER_REPAIR:
      vprint("Training initial model.",self.verbose)
      model = model_factory.build(train_set)
  
      # Check the quality of the initial model on verbose runs.
      if self.verbose:
        print "Calculating original model statistics on test data:"
        print "\tTraining Set:"
        train_pred_tuples = model.test(train_set)
        train_conf_matrix = get_conf_matrix(train_pred_tuples)
        print "\t\tConf-Matrix:", train_conf_matrix
        for measurer in self.measurers:
          print "\t\t{}: {}".format(measurer.__name__, measurer(train_conf_matrix))
  
        print "\tTesting Set:"
        test_pred_tuples = model.test(test_set)
        test_conf_matrix = get_conf_matrix(test_pred_tuples)
        print "\t\tConf-Matrix", test_conf_matrix
        for measurer in self.measurers:
          print "\t\t{}: {}".format(measurer.__name__, measurer(test_conf_matrix))
  
  
      model_or_factory = model
    else:
      model_or_factory = model_factory
  
    # Translate the headers into indexes for the auditor.
    audit_indices_to_ignore = [headers.index(f) for f in features_to_ignore]

    # Don't audit the response feature.
    audit_indices_to_ignore.append(headers.index(response_header))

   # Prepare the auditor.
    auditor = GradientFeatureAuditor(model_or_factory, headers, train_set, test_set,
                                     repair_steps=self.REPAIR_STEPS, kdd=self.kdd,
                                     features_to_ignore=audit_indices_to_ignore,
                                     features_to_audit=features_to_audit,
                                     output_dir=output_dir,dump_all=dump_all)
  
    # Perform the Gradient Feature Audit and dump the audit results into files.
    audit_filenames = auditor.audit(verbose=self.verbose)
 
    ranked_features = []
    for measurer in self.measurers:
      vprint("Ranking audit files by {}.".format(measurer.__name__),self.verbose)
      ranked_graph_filename = "{}/{}.png".format(auditor.OUTPUT_DIR, measurer.__name__)
      graph_audits(audit_filenames, measurer, ranked_graph_filename)
      ranks = rank_audit_files(audit_filenames, measurer)
      vprint("\t{}".format(ranks), self.verbose)
      ranked_features.append( (measurer, ranks) )

    end_time = datetime.now()
    
    # Store a summary of this experiment.
    summary = [
      "Audit Start Time: {}\n".format(start_time),
      "Audit End Time: {}\n".format(end_time),
      "Retrained Per Repair: {}\n".format(self.RETRAIN_MODEL_PER_REPAIR),
      "Model Factory ID: {}\n".format(model_factory.factory_name),
      "Model Type: {}\n".format(model_factory.verbose_factory_name),
      "Non-standard Model Options: {}\n".format(self.model_options),
      "Train Size: {}\n".format(len(train_set)),
      "Test Size: {}\n".format(len(test_set)),
      "Non-standard Ignored Features: {}\n".format(features_to_ignore),
      "Features: {}\n\n".format(headers)]

    # Print summary 
    for line in summary:
      print(line) 

    for ranker, ranks in ranked_features:
      print("Ranked Features by {}: {}\n".format(ranker.__name__, ranks))
      groups = group_audit_ranks(audit_filenames, ranker)
      print("\tApprox. Trend Groups: {}\n".format(groups))

    # Dump all experiment results if opted  
    if dump_all:
      vprint("Dumping original training data.", self.verbose)
      # Dump the train data to the log.
      train_dump = "{}/original_train_data".format(auditor.OUTPUT_DIR)
      with open(train_dump + ".csv", "w") as f:
        writer = csv.writer(f)
        writer.writerow(headers)
        for row in train_set:
          writer.writerow(row)
  
      if self.WRITE_ORIGINAL_PREDICTIONS:
        # Dump the predictions on the test data.
        with open(train_dump + ".predictions", "w") as f:
          writer = csv.writer(f)
          file_headers = ["Response", "Prediction"]
          writer.writerow(file_headers)
          for response, guess in train_pred_tuples:
            writer.writerow([response, guess])
  
      vprint("Dumping original testing data.", self.verbose)
      # Dump the train data to the log.
      test_dump = "{}/original_test_data".format(auditor.OUTPUT_DIR)
      with open(test_dump + ".csv", "w") as f:
        writer = csv.writer(f)
        writer.writerow(headers)
        for row in test_set:
          writer.writerow(row)
  
      if self.WRITE_ORIGINAL_PREDICTIONS:
        # Dump the predictions on the test data.
        with open(test_dump + ".predictions", "w") as f:
          writer = csv.writer(f)
          file_headers = ["Response", "Prediction"]
          writer.writerow(file_headers)
          for response, guess in test_pred_tuples:
            writer.writerow([response, guess])
  
      # Graph ://github.com/algofairness/BlackBoxAuditing.gitthe audit files.
      vprint("Graphing audit files.",self.verbose)
      for audit_filename in audit_filenames:
        audit_image_filename = audit_filename + ".png"
        graph_audit(audit_filename, self.measurers, audit_image_filename)
 
      # Store a graph of how many predictions change as features are repaired.
      vprint("Graphing prediction changes throughout repair.",self.verbose)
      output_image = auditor.OUTPUT_DIR + "/similarity_to_original_predictions.png"
      graph_prediction_consistency(auditor.OUTPUT_DIR, output_image)
  
      # Store a summary of this experiment to file.
      summary_file = "{}/summary.txt".format(auditor.OUTPUT_DIR)
      with open(summary_file, "w") as f:
        for line in summary:
          f.write(line)
  
        for ranker, ranks in ranked_features:
          f.write("Ranked Features by {}: {}\n".format(ranker.__name__, ranks))
          groups = group_audit_ranks(audit_filenames, ranker)
          f.write("\tApprox. Trend Groups: {}\n".format(groups))
  
      vprint("Summary file written to: {}".format(summary_file), self.verbose)

def main():
  # format data
  data = load_data("german")

  # set the auditor
  auditor = Auditor()
  auditor.model = Weka_SVM

  # call the auditor
  auditor(data, output_dir="try", features_to_audit=["checking_status","duration"], dump_all=True)

if __name__=="__main__":
  main() 



import os
import subprocess

def test():
  here = os.path.abspath(os.path.dirname(__file__))
  os.chdir(here)
  print("Running testsuite for BlackBoxAuditing ...")
  output = subprocess.check_output("./run_test_suite.sh")
  print(output)
